#!/usr/bin/env python
# ******************************************************************************
# MIT License
#
# Copyright (c) 2017 Bradley Kenstler
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ******************************************************************************
"""
Cyclical learning rate policy module.

Originated from https://github.com/bckenstler/CLR.
"""

from keras import backend as K
from keras.callbacks import Callback

import numpy as np


class CyclicLR(Callback):
    """This callback implements a cyclical learning rate policy (CLR).
    The method cycles the learning rate between two boundaries with
    some constant frequency, as detailed in this paper (https://arxiv.org/abs/1506.01186).
    The amplitude of the cycle can be scaled on a per-iteration or
    per-cycle basis.
    This class has three built-in policies, as put forth in the paper.
    "triangular":
        A basic triangular cycle w/ no amplitude scaling.
    "triangular2":
        A basic triangular cycle that scales initial amplitude by half each
        cycle.
    "exp_range":
        A cycle that scales initial amplitude by gamma**(cycle iterations) at
        each cycle iteration.
    For more detail, please see paper.

    # Example
        ```python
            clr = CyclicLR(base_lr=0.001, max_lr=0.006,
                                step_size=2000., mode='triangular')
            model.fit(X_train, Y_train, callbacks=[clr])
        ```

    Class also supports custom scaling functions:
        ```python
            clr_fn = lambda x: 0.5*(1+np.sin(x*np.pi/2.))
            clr = CyclicLR(base_lr=0.001, max_lr=0.006,
                                step_size=2000., scale_fn=clr_fn,
                                scale_mode='cycle')
            model.fit(X_train, Y_train, callbacks=[clr])
        ```
    Args:
        base_lr: initial learning rate which is the
            lower boundary in the cycle.
        max_lr: upper boundary in the cycle. Functionally,
            it defines the cycle amplitude (max_lr - base_lr).
            The lr at any cycle is the sum of base_lr
            and some scaling of the amplitude; therefore
            max_lr may not actually be reached depending on
            scaling function.
        step_size: number of training iterations per
            half cycle. Authors suggest setting step_size
            2-8 x training iterations in epoch.
        mode: one of {triangular, triangular2, exp_range}.
            Default 'triangular'.
            Values correspond to policies detailed above.
            If scale_fn is not None, this argument is ignored.
        gamma: constant in 'exp_range' scaling function:
            gamma**(cycle iterations)
        scale_fn: Custom scaling policy defined by a single
            argument lambda function, where
            0 <= scale_fn(x) <= 1 for all x >= 0.
            mode paramater is ignored
        scale_mode: {'cycle', 'iterations'}.
            Defines whether scale_fn is evaluated on
            cycle number or cycle iterations (training
            iterations since start of cycle). Default is 'cycle'.
    """

    def __init__(self,
                 base_lr=0.001,
                 max_lr=0.006,
                 step_size=2000.,
                 mode='triangular',
                 gamma=1.,
                 scale_fn=None,
                 scale_mode='cycle'):
        super().__init__()

        self.base_lr = base_lr
        self.max_lr = max_lr
        self.step_size = step_size
        self.mode = mode
        self.gamma = gamma
        if scale_fn is None:
            if self.mode == 'triangular':
                self.scale_fn = lambda x: 1.
                self.scale_mode = 'cycle'
            elif self.mode == 'triangular2':
                self.scale_fn = lambda x: 1 / (2.**(x - 1))
                self.scale_mode = 'cycle'
            elif self.mode == 'exp_range':
                self.scale_fn = lambda x: gamma**(x)
                self.scale_mode = 'iterations'
        else:
            self.scale_fn = scale_fn
            self.scale_mode = scale_mode
        self.clr_iterations = 0.
        self.trn_iterations = 0.
        self.history = {}

        self._reset()

    def _reset(self, new_base_lr=None, new_max_lr=None, new_step_size=None):
        """Resets cycle iterations.
        Optional boundary/step size adjustment.
        """
        if new_base_lr is not None:
            self.base_lr = new_base_lr
        if new_max_lr is not None:
            self.max_lr = new_max_lr
        if new_step_size is not None:
            self.step_size = new_step_size
        self.clr_iterations = 0.

    def clr(self):
        """ Computes the next learning rate value.

        Returns:
            float: learning rate value.
        """
        cycle = np.floor(1 + self.clr_iterations / (2 * self.step_size))
        x = np.abs(self.clr_iterations / self.step_size - 2 * cycle + 1)
        if self.scale_mode == 'cycle':
            return self.base_lr + (self.max_lr - self.base_lr) * np.maximum(
                0, (1 - x)) * self.scale_fn(cycle)
        return self.base_lr + (self.max_lr - self.base_lr) * np.maximum(
            0, (1 - x)) * self.scale_fn(self.clr_iterations)

    def on_train_begin(self, logs=None):
        logs = logs or {}

        if self.clr_iterations == 0:
            K.set_value(self.model.optimizer.lr, self.base_lr)
        else:
            K.set_value(self.model.optimizer.lr, self.clr())

    def on_batch_end(self, _epoch, logs=None):

        logs = logs or {}
        self.trn_iterations += 1
        self.clr_iterations += 1

        self.history.setdefault('lr',
                                []).append(K.get_value(self.model.optimizer.lr))
        self.history.setdefault('iterations', []).append(self.trn_iterations)

        for k, v in logs.items():
            self.history.setdefault(k, []).append(v)

        K.set_value(self.model.optimizer.lr, self.clr())



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
YOLO model definition for detection.
"""

import pickle

import numpy as np

from keras import Model
from keras.utils.data_utils import get_file

from cnn2snn import load_quantized_model, quantize

from ..layer_blocks import separable_conv_block
from ..imagenet.model_akidanet import akidanet_imagenet


def yolo_base(input_shape=(224, 224, 3),
              classes=1,
              nb_box=5,
              alpha=1.0,
              weight_quantization=0,
              activ_quantization=0,
              input_weight_quantization=None,
              input_scaling=(127.5, -1)):
    """ Instantiates the YOLOv2 architecture.

    Args:
        input_shape (tuple): input shape tuple
        classes (int): number of classes to classify images into
        nb_box (int): number of anchors boxes to use
        alpha (float): controls the width of the model
        weight_quantization (int): sets all weights in the model to have
            a particular quantization bitwidth except for the weights in the
            first layer.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization: sets all activations in the model to have a
            particular activation quantization bitwidth.

            * '0' implements floating point 32-bit activations.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_weight_quantization: sets weight quantization in the first layer.
            Defaults to weight_quantization value.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_scaling (tuple, optional): scale factor and offset to apply to
            inputs. Defaults to (127.5, -1). Note that following Akida
            convention, the scale factor is a number used as a divider.

    Returns:
        keras.Model: a Keras Model instance.

    """
    # Overrides input weight quantization if None
    if input_weight_quantization is None:
        input_weight_quantization = weight_quantization

    # Create an AkidaNet network without top layers
    base_model = akidanet_imagenet(
        input_shape=input_shape,
        alpha=alpha,
        include_top=False,
        weight_quantization=weight_quantization,
        activ_quantization=activ_quantization,
        input_weight_quantization=input_weight_quantization,
        input_scaling=input_scaling)

    # Add YOLO top layers to the base model
    input_shape = base_model.input_shape
    x = base_model.layers[-1].output
    x = separable_conv_block(x,
                             filters=1024,
                             name='1conv',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_activation=True,
                             add_batchnorm=True)
    x = separable_conv_block(x,
                             filters=1024,
                             name='2conv',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_activation=True,
                             add_batchnorm=True)
    x = separable_conv_block(x,
                             filters=1024,
                             name='3conv',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_activation=True,
                             add_batchnorm=True)
    x = separable_conv_block(x,
                             filters=(nb_box * (4 + 1 + classes)),
                             name='detection_layer',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=True,
                             add_activation=False,
                             add_batchnorm=False)

    model = Model(inputs=base_model.input, outputs=x, name='yolo_base')

    # Initialize detection layer weights
    layer = model.get_layer("detection_layer")
    detection_weights = layer.get_weights()

    mu, sigma = 0, 0.1

    grid_size = model.output_shape[1:3]
    grid_area = grid_size[0] * grid_size[1]

    dw_kernel = np.random.normal(mu, sigma,
                                 size=detection_weights[0].shape) / grid_area
    pw_kernel = np.random.normal(mu, sigma,
                                 size=detection_weights[1].shape) / grid_area
    bias = np.random.normal(mu, sigma,
                            size=detection_weights[2].shape) / grid_area

    layer.set_weights([dw_kernel, pw_kernel, bias])

    if ((weight_quantization != 0) or (activ_quantization != 0) or
        (input_weight_quantization != 0)):
        return quantize(model, weight_quantization, activ_quantization,
                        input_weight_quantization)
    return model


def yolo_widerface_pretrained():
    """
    Helper method to retrieve a `yolo_base` model that was trained on WiderFace
    dataset and the anchors that are needed to interpet the model output.

    Returns:
        keras.Model, list: a Keras Model instance and a list of anchors.

    """
    anchors_name = 'widerface_anchors.pkl'
    anchors_path = get_file(
        fname=anchors_name,
        origin='http://data.brainchip.com/dataset-mirror/widerface/' +
        anchors_name,
        cache_subdir='datasets/widerface')
    with open(anchors_path, 'rb') as handle:
        anchors = pickle.load(handle)

    model_name = 'yolo_akidanet_widerface_iq8_wq4_aq4.h5'
    file_hash = '4effe9b2622ef3de2763f869f14d4d18b82f1bb6b69ba4c601aaa0a424f77864'
    model_path = get_file(fname=model_name,
                          origin='http://data.brainchip.com/models/yolo/' +
                          model_name,
                          file_hash=file_hash,
                          cache_subdir='models')

    return load_quantized_model(model_path), anchors


def yolo_voc_pretrained():
    """
    Helper method to retrieve a `yolo_base` model that was trained on PASCAL
    VOC2012 dataset for 'person' and 'car' classes only, and the anchors that
    are needed to interpet the model output.

    Returns:
        keras.Model, list: a Keras Model instance and a list of anchors.

    """
    anchors_name = 'voc_anchors.pkl'
    anchors_path = get_file(
        fname=anchors_name,
        origin='http://data.brainchip.com/dataset-mirror/voc/' + anchors_name,
        cache_subdir='datasets/voc')
    with open(anchors_path, 'rb') as handle:
        anchors = pickle.load(handle)

    model_name = 'yolo_akidanet_voc_iq8_wq4_aq4.h5'
    file_hash = '095861adc0e351ef06ab642c0ef8d651faf6a9f23d8277f96d2b01e1899f9c56'
    model_path = get_file(fname=model_name,
                          origin='http://data.brainchip.com/models/yolo/' +
                          model_name,
                          file_hash=file_hash,
                          cache_subdir='models')

    return load_quantized_model(model_path), anchors



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Filter pruning tools to automatically prune a model"""

import numpy as np
from keras import Sequential, Input
from keras.layers import (InputLayer, Conv2D, SeparableConv2D, Dense,
                          BatchNormalization, MaxPool2D, GlobalAvgPool2D,
                          Dropout, ReLU, Softmax, Activation, Flatten, Reshape,
                          Rescaling)

import cnn2snn


def delete_filters(model, layer_to_prune, filters_to_prune):
    """Deletes filters in the given layer and updates weights in it and its
    subsequent layers.

    A pruned model is returned. Only linear models are supported.

    Args:
        model (keras.Model): the model to prune.
        layer_to_prune (str): the name of the neural layer where filters will be
            deleted.
        filters_to_prune (list): indices of filters to delete in the given
            layer.

    Returns:
        keras.Sequential: the pruned model
    """

    pruned_model = _instantiate_pruned_model(model, layer_to_prune,
                                             len(filters_to_prune))
    _assign_pruned_weights(model, pruned_model, layer_to_prune,
                           filters_to_prune)

    return pruned_model


def _instantiate_pruned_model(model, layer_to_prune, num_filters_to_prune):
    """Returns the pruned model where the layer to prune has less filters than
    the original layer. Note that the weights are not transferred yet
    from the original model to the pruned one.

    Args:
        model (keras.Model): the model to prune.
        layer_to_prune (str): the name of the layer where filters will be
            deleted.
        num_filters_to_prune (int): number of filters to delete in the given
            layer.

    Returns:
        keras.Sequential: the pruned model (without weights from the original
            model)
    """

    pruned_model = Sequential()
    pruned_model.add(Input(model.input_shape[1:]))

    # Create the pruned model by browsing the layers of the original model and
    # by deleting filters/channels in the concerned layers:
    # - for each layer of the original model, create a new layer with the new
    #   configuration
    # - add the new layer to the pruned Sequential model
    for layer in model.layers:

        if isinstance(layer, InputLayer):
            continue

        if isinstance(layer, (Conv2D, SeparableConv2D)):
            new_config = layer.get_config()
            if layer.name == layer_to_prune:
                new_config[
                    'filters'] = new_config['filters'] - num_filters_to_prune
                if new_config['filters'] == 0:
                    raise RuntimeError("Impossible to prune all filters of "
                                       f"layer {layer_to_prune}.")
            new_layer = layer.__class__.from_config(new_config)

        elif isinstance(layer, Dense):
            new_config = layer.get_config()
            if layer.name == layer_to_prune:
                new_config['units'] = new_config['units'] - num_filters_to_prune
                if new_config['units'] == 0:
                    raise RuntimeError("Impossible to prune all units of "
                                       f"layer {layer_to_prune}.")
            new_layer = layer.__class__.from_config(new_config)

        elif isinstance(layer, Reshape):
            new_config = layer.get_config()
            # If the layer preceding a Reshape layer was pruned, the new Reshape
            # layer must handle the new target shape. Note that, in cnn2snn,
            # only Reshape layer (N,)->(1,1,N) is supported.
            if layer.target_shape[-1] != pruned_model.output_shape[-1]:
                new_target_shape = layer.target_shape[:-1] + (
                    pruned_model.output_shape[-1],)
                new_config['target_shape'] = new_target_shape
            new_layer = layer.__class__.from_config(new_config)

        elif isinstance(layer, (BatchNormalization, MaxPool2D, GlobalAvgPool2D,
                                Dropout, Flatten, ReLU, Softmax, Activation,
                                Rescaling, cnn2snn.QuantizedActivation)):
            new_layer = layer.__class__.from_config(layer.get_config())

        else:
            raise RuntimeError(
                f"Layer {layer.name} of type {layer.__class__.__name__} "
                "is not supported for pruning")

        pruned_model.add(new_layer)

    return pruned_model


def _assign_pruned_weights(model, pruned_model, layer_to_prune,
                           filters_to_prune):
    """Assigns new weights from the original model to the pruned one.

    For each layer in the model, two steps are performed:
        - update the outbound mask. The outbound mask (of length the number of
          output channels) represents the filters/channels to keep.
        - set the new pruned weights, according to the inbound/outbound masks.

    Args:
        model (keras.Model): the model to prune.
        pruned_model (keras.Sequential): the pruned model where the weights
            must be set.
        layer_to_prune (str): the name of the layer where filters were deleted.
        filters_to_prune (list): indices of filters to delete in the given
            layer.
    """

    # Mask representing the filters to keep (False = filters to prune)
    inbound_mask = [True] * model.input_shape[-1]

    for layer in model.layers:

        # Step 1: Update the outbound mask.
        # Only layers that modifies the number of channels are concerned (i.e.
        # neural layers and Flatten layers). The mask for the layer to prune has
        # "False" values for indices to prune. The mask for other neural layers
        # is "True" for all output channels.
        if layer.name == layer_to_prune:
            outbound_mask = [
                i not in filters_to_prune for i in range(layer.output_shape[-1])
            ]

        elif isinstance(layer, (Conv2D, SeparableConv2D, Dense)):
            # All other (i.e. not pruned) neural layers reset the outbound_mask
            # (all filters are kept)
            outbound_mask = [True] * layer.output_shape[-1]

        elif isinstance(layer, Flatten):
            # A Flatten layer modifies the number of channels: C_out = H*W*C_in.
            # The outbound mask is updated, depending on the input size:
            # input_size = H*W = C_out/C_in
            input_size = int(layer.output_shape[-1] / layer.input_shape[-1])
            outbound_mask = inbound_mask * input_size

        elif isinstance(layer,
                        (InputLayer, BatchNormalization, MaxPool2D,
                         GlobalAvgPool2D, Dropout, ReLU, Softmax, Activation,
                         Reshape, Rescaling, cnn2snn.QuantizedActivation)):
            # All these layers have no weights and do not change the number of
            # channels. Reshape could modify the number of channels. However we
            # only support (N,) -> (1,1,N), thus there is no change in the
            # number of channels.
            outbound_mask = inbound_mask

        else:
            raise RuntimeError(
                f"Layer {layer.name} of type {layer.__class__.__name__} "
                "is not supported for pruning")

        # Step 2: set new pruned weights.
        # The concerned layers are all layers between the neural layer to prune
        # and the next neural layer. Layers with weights are
        # (Quantized)Conv2D/SeparableConv2D/Dense and BatchNormalization layers.
        if isinstance(layer, Conv2D):
            new_weights = layer.get_weights()
            new_weights[0] = new_weights[0][:, :, inbound_mask, :]
            new_weights[0] = new_weights[0][..., outbound_mask]
            if layer.use_bias:
                new_weights[1] = new_weights[1][outbound_mask]
            pruned_model.get_layer(layer.name).set_weights(new_weights)

        elif isinstance(layer, SeparableConv2D):
            new_weights = layer.get_weights()
            new_weights[0] = new_weights[0][:, :, inbound_mask, :]
            new_weights[1] = new_weights[1][:, :, inbound_mask, :]
            new_weights[1] = new_weights[1][..., outbound_mask]
            if layer.use_bias:
                new_weights[2] = new_weights[2][outbound_mask]
            pruned_model.get_layer(layer.name).set_weights(new_weights)

        elif isinstance(layer, Dense):
            new_weights = layer.get_weights()
            new_weights[0] = new_weights[0][inbound_mask, :]
            new_weights[0] = new_weights[0][:, outbound_mask]
            if layer.use_bias:
                new_weights[1] = new_weights[1][outbound_mask]
            pruned_model.get_layer(layer.name).set_weights(new_weights)

        elif isinstance(layer, BatchNormalization):
            outbound_mask = inbound_mask
            new_weights = layer.get_weights()
            for i in range(4):
                new_weights[i] = new_weights[i][inbound_mask]
            pruned_model.get_layer(layer.name).set_weights(new_weights)

        # Update inbound mask for next layer
        inbound_mask = outbound_mask


def neural_layers(model):
    """ Returns the list of neural layers names, except the last one.

    Args:
        model (keras.Model): the model to prune.

    Returns:
        list: the names of the prunable layers.
    """
    layers = [
        l.name
        for l in model.layers
        if type(l) in (Conv2D, SeparableConv2D, Dense)
    ]
    return layers[:-1]


def _get_weights_for_pruning(layer):
    """Returns the kernel of the layer where filters will be pruned.
    For SeparableConv2D, it returns the pointwise kernel.

    Args:
        layer (keras.Layer): the layer to prune.

    Returns:
        `:obj:np.ndarray`: the weights used for pruning.
    """
    weight_index_per_layer_type = {Conv2D: 0, SeparableConv2D: 1, Dense: 0}
    if type(layer) not in weight_index_per_layer_type:
        raise RuntimeError(
            f"Layer {layer.name} must be of type Conv2D, SeparableConv2D or "
            f"Dense. Receives type {type(layer)}.")
    weight_index = weight_index_per_layer_type[type(layer)]
    return layer.get_weights()[weight_index]


def _filter_magnitudes(layer):
    """Returns the absolute mean of filters for the given layer. This is called
    the magnitude of a filter.

    Args:
        layer (keras.Layer): the layer to prune.

    Returns:
        `:obj:np.ndarray`: the magnitudes of the layer's filters.
    """
    weight = _get_weights_for_pruning(layer)
    return np.mean(np.abs(weight), axis=tuple(range(weight.ndim - 1)))


def smallest_filters(layer, pruning_rate):
    """Returns indices of the smallest filters to prune, based on the rate
    of filters to prune.
    Pruning rate must be a value between 0 and 1. For instance, 0.3 means
    that 30% of the filters will be pruned. The function then returns the
    indices of the 30% smallest filters in this layer.

    Args:
        layer (keras.Layer): the layer to prune.
        pruning_rate (float): the pruning rate (between 0 and 1).

    Returns:
        `:obj:np.ndarray`: the indices of the filters to prune.
    """
    l1_norms = _filter_magnitudes(layer)
    indices_sorted = np.argsort(l1_norms)
    num_filters_to_prune = int(np.round(l1_norms.size * pruning_rate))
    return indices_sorted[:num_filters_to_prune]


def prune_model(model,
                acceptance_function,
                pruning_rates=None,
                prunable_layers_policy=neural_layers,
                prunable_filters_policy=smallest_filters):
    """Prune model automatically based on an acceptance function.

    The algorithm for filter pruning is as follows:

        1. Select the first prunable layer (according to the
           ``prunable_layers_policy`` function).
        2. As long as the ``acceptance_function`` returns True, prune
           successively the layer with different pruning rates (according to
           ``pruning_rates`` and ``prunable_filters_policy``).
        3. When the current pruned model is not acceptable, the last valid
           pruning rate is selected for the final pruned model.
        4. Repeat steps 1, 2 and 3 for the next prunable layers.


    Examples:

        .. code-block:: python

            acceptable_drop = 0.05

            def acceptance_function(base_model, pruned_model):
                # This function returns True if the pruned_model is acceptable.
                # Here, the pruned model is acceptable if the accuracy drops
                # less than 5% from the base model.

                def evaluate(model):
                    _, accuracy = model.evaluate(data, labels)
                    return accuracy

                return evaluate(base_model) - evaluate(pruned_model) <= acceptable_drop

            # Prune model
            pruned_model, pruning_rates = prune_model(model, acceptance_function)

    Args:
        model (keras.Model): a keras model to prune
        acceptance_function (function): a criterion function that returns True
            if the pruned model is acceptable. The signature must be
            `function(base_model, pruned_model)`.
        pruning_rates (list, optional): a list of pruning rates to test. Default
            is [0.1, 0.2, ..., 0.9].
        prunable_layers_policy (function, optional): a function returning the
            layers to prune in the model. The signature must be
            `function(model)`, and must return a list of prunable layer names.
            By default, all neural layers (Conv2D/SeparableConv2D/Dense) are
            candidates for pruning.
        prunable_filters_policy (function, optional): a function that returns
            the filters to prune in a given layer for a specific pruning rate.
            The signature must be `function(layer, pruning_rate)` and returns a
            list of indices to prune. By default, filters with the lowest
            magnitude are pruned.

    Returns:
        tuple: the pruned model and the pruning rates.
    """

    if pruning_rates is None:
        pruning_rates = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]

    # Start by creating a copy of the model where batchnorm is folded
    pruned_model = cnn2snn.transforms.sequentialize(model)
    pruned_model = cnn2snn.transforms.invert_batchnorm_pooling(pruned_model)
    pruned_model = cnn2snn.transforms.fold_batchnorm(pruned_model)

    # Get the list of prunable layers
    layers_to_prune = prunable_layers_policy(pruned_model)

    # Initialize pruning rates (no pruning)
    rates = {name: 0.0 for name in layers_to_prune}

    for name in layers_to_prune:
        layer = pruned_model.get_layer(name)
        next_pruned_model = None

        for pruning_rate in pruning_rates:
            # Get the list of prunable filters for the specified rate
            filters = prunable_filters_policy(layer, pruning_rate)
            candidate_model = delete_filters(pruned_model, layer.name, filters)

            if acceptance_function(pruned_model, candidate_model):
                # The candidate model passed the acceptance criteria
                next_pruned_model = candidate_model
                rates[name] = pruning_rate
            else:
                # Stop at that rate
                break

        if next_pruned_model:
            pruned_model = next_pruned_model

    return pruned_model, rates



# ******************************************************************************
#!/usr/bin/env python
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Training script for YOLO models.
"""

import pickle
import argparse

from keras import Model
from keras.layers import Reshape
from keras.callbacks import EarlyStopping

from cnn2snn import load_quantized_model, convert

from .yolo_loss import YoloLoss
from .map_evaluation import MapEvaluation
from .batch_generator import BatchGenerator
from ..training import get_training_parser, freeze_model_before, compile_model


def get_data(data_path, anchors_path):
    """ Loads data.

    Args:
        data_path (str): path to data pickle file
        anchors_path (str): path to anchors pickle file

    Returns:
        dict, dict, list, list: train and validation data, labels and anchors
    """
    with open(data_path, 'rb') as handle:
        train_data, valid_data, labels = pickle.load(handle)

    with open(anchors_path, 'rb') as handle:
        anchors = pickle.load(handle)

    return train_data, valid_data, labels, anchors


def train(model, train_data, valid_data, anchors, labels, obj_threshold,
          nms_threshold, epochs, batch_size, grid_size):
    """ Trains the model.

    Args:
        model (keras.Model): the model to train
        train_data (dict): training data
        valid_data (dict): validation data
        anchors (list): list of anchors
        labels (list): list of labels
        obj_threshold (float): confidence threshold for a box
        nms_threshold (float): non-maximal supression threshold
        epochs (int): the number of epochs
        batch_size (int): the batch size
        grid_size (tuple): the grid size
    """
    TRAIN_TIMES = 10

    # Build batch generators
    input_shape = model.input.shape[1:]

    train_generator = BatchGenerator(input_shape=input_shape,
                                     data=train_data,
                                     grid_size=grid_size,
                                     labels=labels,
                                     anchors=anchors,
                                     batch_size=batch_size)

    valid_generator = BatchGenerator(input_shape=input_shape,
                                     data=valid_data,
                                     grid_size=grid_size,
                                     labels=labels,
                                     anchors=anchors,
                                     batch_size=batch_size,
                                     jitter=False)

    # Create callbacks
    early_stop_cb = EarlyStopping(monitor='val_loss',
                                  min_delta=0.001,
                                  patience=10,
                                  mode='min',
                                  verbose=1)

    map_evaluator_cb = MapEvaluation(model,
                                     valid_data,
                                     labels,
                                     anchors,
                                     period=4,
                                     obj_threshold=obj_threshold,
                                     nms_threshold=nms_threshold)

    callbacks = [early_stop_cb, map_evaluator_cb]

    # Start the training process
    model.fit(x=train_generator,
              steps_per_epoch=len(train_generator) * TRAIN_TIMES,
              epochs=epochs,
              validation_data=valid_generator,
              callbacks=callbacks,
              workers=12,
              max_queue_size=40)


def evaluate(model, valid_data, anchors, labels, obj_threshold, nms_threshold):
    """ Evaluates model performances.

    Args:
        model (keras.Model or akida.Model): the model to evaluate
        valid_data (dict): evaluation data
        anchors (list): list of anchors
        labels (list): list of labels
        obj_threshold (float): confidence threshold for a box
        nms_threshold (float): non-maximal supression threshold
    """
    # Create the mAP evaluator
    map_evaluator = MapEvaluation(model,
                                  valid_data,
                                  labels,
                                  anchors,
                                  obj_threshold=obj_threshold,
                                  nms_threshold=nms_threshold,
                                  is_keras_model=isinstance(model, Model))

    # Compute mAP scores and display them
    mAP, average_precisions = map_evaluator.evaluate_map()
    for label, average_precision in average_precisions.items():
        print(labels[label], '{:.4f}'.format(average_precision))
    print('mAP: {:.4f}'.format(mAP))


def main():
    """ Entry point for script and CLI usage.
    """
    global_parser = argparse.ArgumentParser(add_help=False)
    global_parser.add_argument(
        "-d",
        "--data",
        default=None,
        help="The pickle file generated by the preprocessing"
        " step that contains data variables")
    global_parser.add_argument("-ap",
                               "--anchors_path",
                               default=None,
                               help="Path to anchors boxes file.")
    global_parser.add_argument("-obj",
                               "--obj_thresh",
                               type=float,
                               default=0.5,
                               help="Confidence threshold for a box")
    global_parser.add_argument("-nms",
                               "--nms_thresh",
                               type=float,
                               default=0.5,
                               help="Non-Maximal Suppression threshold.")

    parser = get_training_parser(batch_size=128,
                                 freeze_before=True,
                                 tune=True,
                                 global_parser=global_parser)[0]
    args = parser.parse_args()

    # Load data
    train_data, valid_data, labels, anchors = get_data(args.data,
                                                       args.anchors_path)

    # Load the source model
    base_model = load_quantized_model(args.model)

    # Create a final reshape layer for loss computation
    grid_size = base_model.output_shape[1:3]
    output = Reshape(
        (grid_size[1], grid_size[0], len(anchors), 4 + 1 + len(labels)),
        name="YOLO_output")(base_model.output)

    # Build the full model
    model = Model(base_model.input, output)

    # Freeze the model
    if "freeze_before" in args:
        freeze_model_before(model, args.freeze_before)

    # Compile model
    learning_rate = 1e-4
    if args.action == "tune":
        learning_rate = 1e-5

    compile_model(model,
                  learning_rate=learning_rate,
                  loss=YoloLoss(anchors, grid_size, args.batch_size),
                  metrics=None)

    # Train model
    if args.action in ["train", "tune"]:
        train(model, train_data, valid_data, anchors, labels, args.obj_thresh,
              args.nms_thresh, args.epochs, args.batch_size, grid_size)

        if args.savemodel:
            # Remove the last reshape layer introduced for training
            model = Model(model.input, model.layers[-2].output)
            # Save the model
            model.save(args.savemodel, include_optimizer=False)
            print(f"Trained model saved as {args.savemodel}")

    elif args.action == 'eval':
        # Evaluate model accuracy
        if args.akida:
            # Drop the last reshape layer that is not Akida compatible
            if model.layers[-1].name == 'YOLO_output':
                model = Model(model.input, model.layers[-2].output)
            model = convert(model)
        evaluate(model, valid_data, anchors, labels, args.obj_thresh,
                 args.nms_thresh)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Training script for UTKFace model.
"""
import numpy as np

from keras.callbacks import LearningRateScheduler

from cnn2snn import load_quantized_model, convert

from ..training import get_training_parser, compile_model, evaluate_model
from .preprocessing import load_data


def get_data():
    """ Loads UTKFace data.

    Returns:
        np.array, np.array, np.array, np.array:  train set, train labels, test
            set and test labels
    """
    # Load the dataset
    x_train, y_train, x_test, y_test = load_data()

    return x_train.astype('float32'), y_train, x_test.astype('float32'), y_test


def train_model(model, x_train, y_train, x_test, y_test, epochs, batch_size):
    """ Trains the model.

    Args:
        model (keras.Model): the model to train
        x_train (numpy.ndarray): train data
        y_train (numpy.ndarray): train labels
        x_test (numpy.ndarray): test data
        y_test (numpy.ndarray): test labels
        epochs (int):  the number of epochs
        batch_size (int): the batch size
    """
    # Learning rate: be more aggressive at the beginning, and apply decay
    lr_start = 1e-3
    lr_end = 1e-4
    lr_decay = (lr_end / lr_start)**(1. / epochs)

    lr_scheduler = LearningRateScheduler(lambda e: lr_start * lr_decay**e)
    callbacks = [lr_scheduler]

    history = model.fit(x_train,
                        y_train,
                        batch_size=batch_size,
                        epochs=epochs,
                        verbose=1,
                        validation_data=(x_test, y_test),
                        callbacks=callbacks)
    print(history.history)


def main():
    """ Entry point for script and CLI usage.
    """
    parser = get_training_parser(batch_size=128, global_batch_size=False)[0]
    args = parser.parse_args()

    # Load the source model
    model = load_quantized_model(args.model)

    # Compile model
    compile_model(model, loss='mae', metrics=None)

    # Load data
    x_train, y_train, x_test, y_test = get_data()

    # Train model
    if args.action == "train":
        train_model(model, x_train, y_train, x_test, y_test, args.epochs,
                    args.batch_size)

        # Save model in Keras format (h5)
        if args.savemodel:
            model.save(args.savemodel, include_optimizer=False)
            print(f"Trained model saved as {args.savemodel}")

    elif args.action == "eval":
        # Evaluate model accuracy
        if args.akida:
            model_ak = convert(model)
            y = model_ak.evaluate(x_test)
            mae = np.sum(np.abs(y_test.squeeze() - y.squeeze())) / len(y_test)
            print("Validation accuracy: {0:.4f}".format(mae))
        else:
            evaluate_model(model, x_test, y=y_test, print_history=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
AkidaNet model definition for ImageNet classification.

AkidaNet is an NSoC optimized model inspired from VGG and MobileNet V1
architectures. It can be used for multiple use cases through transfer learning.

"""

import warnings

from keras import Model, regularizers
from keras.layers import (Input, Reshape, Dropout, Activation, Rescaling)
from keras.utils.data_utils import get_file

from cnn2snn import quantize, load_quantized_model

from ..layer_blocks import conv_block, separable_conv_block

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/akidanet/'


def _obtain_input_shape(input_shape, default_size, min_size, include_top):
    """Internal utility to compute/validate a model's input shape.

    Args:
        input_shape: either None (will return the default model input shape),
            or a user-provided shape to be validated.
        default_size: default input width/height for the model.
        min_size: minimum input width/height accepted by the model.
        include_top: whether the model is expected to
            be linked to a classifier via a Flatten layer.

    Returns:
        tuple of integers: input shape (may include None entries).

    Raises:
        ValueError: in case of invalid argument values.
    """
    if input_shape and len(input_shape) == 3:
        if input_shape[-1] not in {1, 3}:
            warnings.warn('This model usually expects 1 or 3 input channels. '
                          'However, it was passed an input_shape with ' +
                          str(input_shape[-1]) + ' input channels.')
        default_shape = (default_size, default_size, input_shape[-1])
    else:
        default_shape = (default_size, default_size, 3)

    if input_shape:
        if input_shape is not None:
            if len(input_shape) != 3:
                raise ValueError(
                    '`input_shape` must be a tuple of three integers.')
            if ((input_shape[0] is not None and input_shape[0] < min_size) or
                (input_shape[1] is not None and input_shape[1] < min_size)):
                raise ValueError('Input size must be at least ' +
                                 str(min_size) + 'x' + str(min_size) +
                                 '; got `input_shape=' + str(input_shape) + '`')
    else:
        if include_top:
            input_shape = default_shape
        else:
            input_shape = (None, None, 3)
    if include_top:
        if None in input_shape:
            raise ValueError('If `include_top` is True, '
                             'you should specify a static `input_shape`. '
                             'Got `input_shape=' + str(input_shape) + '`')
    return input_shape


def akidanet_imagenet(input_shape=None,
                      alpha=1.0,
                      include_top=True,
                      pooling=None,
                      classes=1000,
                      weight_quantization=0,
                      activ_quantization=0,
                      input_weight_quantization=None,
                      input_scaling=(128, -1)):
    """Instantiates the AkidaNet architecture.

    Args:
        input_shape (tuple): optional shape tuple.
        alpha (float): controls the width of the model.

            * If `alpha` < 1.0, proportionally decreases the number of filters
              in each layer.
            * If `alpha` > 1.0, proportionally increases the number of filters
              in each layer.
            * If `alpha` = 1, default number of filters from the paper are used
              at each layer.
        include_top (bool): whether to include the fully-connected
            layer at the top of the model.
        pooling (str): optional pooling mode for feature extraction
            when `include_top` is `False`.

            * `None` means that the output of the model will be the 4D tensor
              output of the last convolutional block.
            * `avg` means that global average pooling will be applied to the
              output of the last convolutional block, and thus the output of the
              model will be a 2D tensor.
        classes (int): optional number of classes to classify images
            into, only to be specified if `include_top` is `True`.
        weight_quantization (int): sets all weights in the model to have
            a particular quantization bitwidth except for the weights in the
            first layer.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization: sets all activations in the model to have a
            particular activation quantization bitwidth.

            * '0' implements floating point 32-bit activations.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_weight_quantization: sets weight quantization in the first layer.
            Defaults to weight_quantization value.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_scaling (tuple, optional): scale factor and offset to apply to
            inputs. Defaults to (128, -1). Note that following Akida convention,
            the scale factor is an integer used as a divider.

    Returns:
        keras.Model: a Keras model for AkidaNet/ImageNet.

    Raises:
        ValueError: in case of invalid input shape.
    """
    # check if overrides have been provided and override
    if input_weight_quantization is None:
        input_weight_quantization = weight_quantization

    # Define weight regularization, will apply to the convolutional layers and
    # to all pointwise weights of separable convolutional layers.
    weight_regularizer = regularizers.l2(4e-5)

    # Determine proper input shape and default size.
    if input_shape is None:
        default_size = 224
    else:
        rows = input_shape[0]
        cols = input_shape[1]

        if rows == cols and rows in [128, 160, 192, 224]:
            default_size = rows
        else:
            default_size = 224

    input_shape = _obtain_input_shape(input_shape,
                                      default_size=default_size,
                                      min_size=32,
                                      include_top=include_top)

    img_input = Input(shape=input_shape)

    if input_scaling is None:
        x = img_input
    else:
        scale, offset = input_scaling
        x = Rescaling(1. / scale, offset)(img_input)

    x = conv_block(x,
                   filters=int(32 * alpha),
                   name='conv_0',
                   kernel_size=(3, 3),
                   padding='same',
                   use_bias=False,
                   strides=2,
                   add_batchnorm=True,
                   add_activation=True,
                   kernel_regularizer=weight_regularizer)

    x = conv_block(x,
                   filters=int(64 * alpha),
                   name='conv_1',
                   kernel_size=(3, 3),
                   padding='same',
                   use_bias=False,
                   add_batchnorm=True,
                   add_activation=True,
                   kernel_regularizer=weight_regularizer)

    x = conv_block(x,
                   filters=int(128 * alpha),
                   name='conv_2',
                   kernel_size=(3, 3),
                   padding='same',
                   strides=2,
                   use_bias=False,
                   add_batchnorm=True,
                   add_activation=True,
                   kernel_regularizer=weight_regularizer)

    x = conv_block(x,
                   filters=int(128 * alpha),
                   name='conv_3',
                   kernel_size=(3, 3),
                   padding='same',
                   use_bias=False,
                   add_batchnorm=True,
                   add_activation=True,
                   kernel_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(256 * alpha),
                             name='separable_4',
                             kernel_size=(3, 3),
                             padding='same',
                             strides=2,
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(256 * alpha),
                             name='separable_5',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_6',
                             kernel_size=(3, 3),
                             padding='same',
                             strides=2,
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_7',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_8',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_9',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_10',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_11',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(1024 * alpha),
                             name='separable_12',
                             kernel_size=(3, 3),
                             padding='same',
                             strides=2,
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    # Last separable layer with global pooling
    layer_pooling = 'global_avg' if include_top or pooling == 'avg' else None
    x = separable_conv_block(x,
                             filters=int(1024 * alpha),
                             name='separable_13',
                             kernel_size=(3, 3),
                             padding='same',
                             pooling=layer_pooling,
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    if include_top:
        shape = (1, 1, int(1024 * alpha))

        x = Reshape(shape, name='reshape_1')(x)
        x = Dropout(1e-3, name='dropout')(x)

        x = separable_conv_block(x,
                                 filters=classes,
                                 name='separable_14',
                                 kernel_size=(3, 3),
                                 padding='same',
                                 use_bias=False,
                                 add_batchnorm=False,
                                 add_activation=False,
                                 pointwise_regularizer=weight_regularizer)
        act_function = 'softmax' if classes > 1 else 'sigmoid'
        x = Activation(act_function, name=f'act_{act_function}')(x)
        x = Reshape((classes,), name='reshape_2')(x)

    # Create model.
    model = Model(img_input,
                  x,
                  name='akidanet_%0.2f_%s_%s' %
                  (alpha, input_shape[0], classes))

    if ((weight_quantization != 0) or (activ_quantization != 0) or
        (input_weight_quantization != 0)):
        return quantize(model, weight_quantization, activ_quantization,
                        input_weight_quantization)
    return model


def akidanet_imagenet_pretrained(alpha=1.0):
    """
    Helper method to retrieve an `akidanet_imagenet` model that was trained on
    ImageNet dataset.

    Args:
        alpha (float, optional): width of the model, allowed values in [0.25,
            0.5, 1]. Defaults to 1.0.

    Returns:
        keras.Model: a Keras Model instance.

    """
    if alpha == 1.0:
        model_name = 'akidanet_imagenet_224_iq8_wq4_aq4.h5'
        file_hash = '1be87e301fe9280d137a821de23fe9ac47c4c7e9ec8405f2279fd0ea02a648c7'
    elif alpha == 0.5:
        model_name = 'akidanet_imagenet_224_alpha_50_iq8_wq4_aq4.h5'
        file_hash = 'bd9ff5d5944d25d220883d2feaff8d34a6cafd3d8a44d98a3307281410781b6e'
    elif alpha == 0.25:
        model_name = 'akidanet_imagenet_224_alpha_25_iq8_wq4_aq4.h5'
        file_hash = '6ee45047b9e283576946d2a65e15411bdc82f1e02faf1b684ed0730089aae69c'
    else:
        raise ValueError(
            f"Requested model with alpha={alpha} is not available.")

    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)


def akidanet_cats_vs_dogs_pretrained():
    """
    Helper method to retrieve an `akidanet_imagenet` model that was trained on
    Cats vs.Dogs dataset.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'akidanet_cats_vs_dogs_iq8_wq4_aq4_lwq_float.h5'
    file_hash = '7f5d71f828c379503ed9d1c1a37e13611369ee558655c6b3406186a05340d35a'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)


def akidanet_imagenette_pretrained(alpha=1.0):
    """
    Helper method to retrieve a `akidanet_imagenet` model that was trained on
    Imagenette dataset.

    Args:
        alpha (float, optional): width of the model, allowed values in [0.25,
            0.5, 1]. Defaults to 1.0.

    Returns:
        keras.Model: a Keras Model instance.
    """
    if alpha == 1.0:
        model_name = 'akidanet_imagenette_224_iq8_wq4_aq4.h5'
        file_hash = 'c99e1249802bb5b6348ca6746416cc9e903ef5c43f52ea734caef15916494904'
    elif alpha == 0.5:
        model_name = 'akidanet_imagenette_224_alpha_50_iq8_wq4_aq4.h5'
        file_hash = '6729a203605f527fd0cff76d76275ef74be2f467cd5e3abc9df3d7df54c1c2c9'
    elif alpha == 0.25:
        model_name = 'akidanet_imagenette_224_alpha_25_iq8_wq4_aq4.h5'
        file_hash = 'c99e1249802bb5b6348ca6746416cc9e903ef5c43f52ea734caef15916494904'
    else:
        raise ValueError(
            f"Requested model with alpha={alpha} is not available.")

    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)


def akidanet_faceidentification_pretrained():
    """
    Helper method to retrieve an `akidanet_imagenet` model that was trained on
    CASIA Webface dataset and that performs face identification.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'akidanet_faceidentification_iq8_wq4_aq4.h5'
    file_hash = 'a50a89a5608c1f01bb9ab70ed214be3e249890bc6a41bc179c2a9427020a2c95'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)


def akidanet_faceverification_pretrained():
    """
    Helper method to retrieve an `akidanet_imagenet` model that was trained on
    CASIA Webface dataset and optimized with ArcFace that can perform face
    verification on LFW.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'akidanet_faceverification_iq8_wq4_aq4.h5'
    file_hash = '8e1c47013dc17745901041f2060aa1e2b5f7222749ec5068b3c9549a42eab070'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)


def akidanet_melanoma_pretrained():
    """
    Helper method to retrieve an `akidanet_imagenet` model that was trained on
    SIIM-ISIC Melanoma Classification dataset.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'akidanet_melanoma_iq8_wq4_aq4.h5'
    file_hash = '844e56a7f5c642504483a19d307a2524d127716bf1d5a6bd4ead389f23b1e99a'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)


def akidanet_odir5k_pretrained():
    """
    Helper method to retrieve an `akidanet_imagenet` model that was trained on
    ODIR-5K dataset.

    The model focuses on the following classes that are a part of the original
    dataset: normal, cataract, AMD (age related macular degeneration) and
    pathological myopia.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'akidanet_odir5k_iq8_wq4_aq4.h5'
    file_hash = 'f8af24f8443191b493c1d82eaf3d8377018d056a936c0e9aea0e4fb701328b09'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)


def akidanet_retinal_oct_pretrained():
    """
    Helper method to retrieve an `akidanet_imagenet` model that was trained on
    retinal OCT dataset.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'akidanet_retinal_oct_iq8_wq4_aq4.h5'
    file_hash = '4811d9ec9ba46434074d3edfae7733557d4f3cde473bb2d194d3c3200f060b64'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)


def akidanet_ecg_pretrained():
    """
    Helper method to retrieve an `akidanet_imagenet` model that was trained on
    ECG classification Physionet2017 dataset.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'akidanet_ecg_iq8_wq4_aq4.h5'
    file_hash = 'd3bf3b598e1e0175873e60f39cd5b6cefa874f4d351518f1a788e402d7bfaa6b'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)


def akidanet_plantvillage_pretrained():
    """
    Helper method to retrieve an `akidanet_imagenet` model that was trained on
    PlantVillage dataset.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'akidanet_plantvillage_iq8_wq4_aq4.h5'
    file_hash = 'beace6e5a23ee690826252124b1da1f523f120cca15b117858adb2ccfd75420d'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2017-2018 Fizyr (https://fizyr.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Module used to compute mAP scores for YOLO classification.
"""

import keras
import numpy as np

from .processing import load_image, preprocess_image, decode_output


class MapEvaluation(keras.callbacks.Callback):
    """ Evaluate a given dataset using a given model.
        Code originally from https://github.com/fizyr/keras-retinanet.

        Args:
            model (keras.Model): model to evaluate.
            val_data (dict): dictionary containing validation data as obtained
             using `preprocess_widerface.py` module
            labels (list): list of labels as strings
            anchors (list): list of anchors boxes
            period (int, optional): periodicity the precision is printed,
             defaults to once per epoch.
            obj_threshold (float, optional): confidence threshold for a box
            nms_threshold (float, optional): non-maximal supression threshold
            max_box_per_image (int, optional): maximum number of detections per
             image
            is_keras_model (bool, optional): indicated if the model is a Keras
             model (True) or an Akida model (False)

        Returns:
            A dict mapping class names to mAP scores.
    """

    def __init__(self,
                 model,
                 val_data,
                 labels,
                 anchors,
                 period=1,
                 obj_threshold=0.5,
                 nms_threshold=0.5,
                 max_box_per_image=10,
                 is_keras_model=True):

        super().__init__()
        self._model = model
        self._data = val_data
        self._data_len = len(val_data)
        self._labels = labels
        self._num_classes = len(labels)
        self._anchors = anchors
        self._period = period
        self._obj_threshold = obj_threshold
        self._nms_threshold = nms_threshold
        self._max_box_per_image = max_box_per_image
        self._is_keras_model = is_keras_model

    def on_epoch_end(self, epoch, logs=None):
        """ Keras callback called at the end of an epoch.

        Args:
            epoch (int): index of epoch.
            logs (dict, optional): metric results for this training epoch, and
                for the validation epoch if validation is performed. Validation
                result keys are prefixed with val. For training epoch, the
                values of the Model’s metrics are returned.
                Example : {‘loss’: 0.2, ‘acc’: 0.7}. Defaults to None.
        """
        if epoch % self._period == 0 and self._period != 0:
            _map, average_precisions = self.evaluate_map()
            print('\n')
            for label, average_precision in average_precisions.items():
                print(self._labels[label], '{:.4f}'.format(average_precision))
            print('mAP: {:.4f}'.format(_map))

    def evaluate_map(self):
        """ Evaluates current mAP score on the model.

        Returns:
            tuple: global mAP score and dictionnary of label and mAP for each
            class.
        """
        average_precisions = self._calc_avg_precisions()
        _map = sum(average_precisions.values()) / len(average_precisions)

        return _map, average_precisions

    def _load_annotations(self, i):
        annots = []

        for obj in self._data[i]['boxes']:
            annot = [
                obj['x1'], obj['y1'], obj['x2'], obj['y2'],
                self._labels.index(obj['label'])
            ]
            annots += [annot]

        return np.array(annots)

    def _calc_avg_precisions(self):

        # gather all detections and annotations
        all_detections = [[None
                           for _ in range(self._num_classes)]
                          for _ in range(self._data_len)]
        all_annotations = [[None
                            for _ in range(self._num_classes)]
                           for _ in range(self._data_len)]

        for i in range(self._data_len):
            raw_image = load_image(self._data[i]['image_path'])
            raw_height, raw_width, _ = raw_image.shape

            if self._is_keras_model:
                image = preprocess_image(raw_image, self._model.input_shape[1:])
                input_image = image[np.newaxis, :]
                output = self._model.predict(input_image)[0]
            else:
                image = preprocess_image(raw_image,
                                         self._model.layers[0].input_dims)
                input_image = image[np.newaxis, :].astype(np.uint8)
                potentials = self._model.evaluate(input_image)[0]

                h, w, _ = potentials.shape
                potentials = potentials.reshape(
                    (h, w, len(self._anchors), 4 + 1 + self._num_classes))
                output = potentials.transpose((1, 0, 2, 3))

            pred_boxes = decode_output(output, self._anchors, self._num_classes,
                                       self._obj_threshold, self._nms_threshold)

            score = np.array([box.get_score() for box in pred_boxes])
            pred_labels = np.array([box.get_label() for box in pred_boxes])

            if len(pred_boxes) > 0:
                pred_boxes = np.array([[
                    box.x1 * raw_width, box.y1 * raw_height, box.x2 * raw_width,
                    box.y2 * raw_height, box.score
                ] for box in pred_boxes])
            else:
                pred_boxes = np.array([[]])

            # sort the boxes and the labels according to scores
            score_sort = np.argsort(-score)
            pred_labels = pred_labels[score_sort]
            pred_boxes = pred_boxes[score_sort]

            # limit the number of predictions to max_box_per_image based on
            # score
            number_of_predictions = pred_boxes.shape[0]
            if number_of_predictions > self._max_box_per_image:
                pred_labels = pred_labels[:self._max_box_per_image]
                pred_boxes = pred_boxes[:self._max_box_per_image, :]

            # copy detections to all_detections
            for label in range(self._num_classes):
                all_detections[i][label] = pred_boxes[pred_labels == label, :]

            annotations = self._load_annotations(i)

            # copy ground truth to all_annotations
            for label in range(self._num_classes):
                all_annotations[i][label] = annotations[annotations[:, 4] ==
                                                        label, :4].copy()

        # compute mAP by comparing all detections and all annotations
        average_precisions = {}

        for label in range(self._num_classes):
            false_positives = np.zeros((0,))
            true_positives = np.zeros((0,))
            scores = np.zeros((0,))
            num_annotations = 0.0

            for i in range(self._data_len):
                detections = all_detections[i][label]
                annotations = all_annotations[i][label]
                num_annotations += annotations.shape[0]
                detected_annotations = []

                for d in detections:
                    scores = np.append(scores, d[4])

                    if annotations.shape[0] == 0:
                        false_positives = np.append(false_positives, 1)
                        true_positives = np.append(true_positives, 0)
                        continue

                    overlaps = self._compute_overlap(np.expand_dims(d, axis=0),
                                                     annotations)
                    assigned_annotation = np.argmax(overlaps, axis=1)
                    max_overlap = overlaps[0, assigned_annotation]

                    if max_overlap >= 0.5 and assigned_annotation not in detected_annotations:
                        false_positives = np.append(false_positives, 0)
                        true_positives = np.append(true_positives, 1)
                        detected_annotations.append(assigned_annotation)
                    else:
                        false_positives = np.append(false_positives, 1)
                        true_positives = np.append(true_positives, 0)

            # no annotations -> AP for this class is 0 (is this correct?)
            if num_annotations == 0:
                average_precisions[label] = 0
                continue

            # sort by score
            indices = np.argsort(-scores)
            false_positives = false_positives[indices]
            true_positives = true_positives[indices]

            # compute false positives and true positives
            false_positives = np.cumsum(false_positives)
            true_positives = np.cumsum(true_positives)

            # compute recall and precision
            recall = true_positives / num_annotations
            precision = true_positives / np.maximum(
                true_positives + false_positives,
                np.finfo(np.float64).eps)

            # compute average precision
            average_precision = self._compute_ap(recall, precision)
            average_precisions[label] = average_precision

        return average_precisions

    @staticmethod
    def _compute_overlap(a, b):
        """ Computes overlap between two boxes defined by their top left and
        bottom right corner coordinates.
        Code originally from https://github.com/rbgirshick/py-faster-rcnn.

        Args:
            a (ndarray): (N, 4) ndarray of float
            b (ndarray): (K, 4) ndarray of float

        Returns:
            (N, K) ndarray of overlap between input boxes
        """
        area = (b[:, 2] - b[:, 0]) * (b[:, 3] - b[:, 1])

        iw = np.minimum(np.expand_dims(a[:, 2], axis=1), b[:, 2]) - np.maximum(
            np.expand_dims(a[:, 0], 1), b[:, 0])
        ih = np.minimum(np.expand_dims(a[:, 3], axis=1), b[:, 3]) - np.maximum(
            np.expand_dims(a[:, 1], 1), b[:, 1])

        iw = np.maximum(iw, 0)
        ih = np.maximum(ih, 0)

        ua = np.expand_dims(
            (a[:, 2] - a[:, 0]) * (a[:, 3] - a[:, 1]), axis=1) + area - iw * ih

        ua = np.maximum(ua, np.finfo(float).eps)

        intersection = iw * ih

        return intersection / ua

    @staticmethod
    def _compute_ap(recall, precision):
        """ Compute the average precision, given the recall and precision
        curves.
        Code originally from https://github.com/rbgirshick/py-faster-rcnn.

        Args:
            recall (list): the recall curve
            precision (list): the precision curve

        Returns:
            The average precision as computed in py-faster-rcnn.
        """
        # correct AP calculation
        # first append sentinel values at the end
        mrec = np.concatenate(([0.], recall, [1.]))
        mpre = np.concatenate(([0.], precision, [0.]))

        # compute the precision envelope
        for i in range(mpre.size - 1, 0, -1):
            mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])

        # to calculate area under PR curve, look for points
        # where X axis (recall) changes value
        i = np.where(mrec[1:] != mrec[:-1])[0]

        # and sum (\Delta recall) * prec
        ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])
        return ap



#!/usr/bin/env python
# *****************************************************************************
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Tools for Knowledge Distillation training.

Originated from https://keras.io/examples/vision/knowledge_distillation/.

Reference Hinton et al. (2015) https://arxiv.org/abs/1503.02531
"""

import tensorflow as tf
from tensorflow import GradientTape
from keras import Model
from keras.losses import KLDivergence


class Distiller(Model):
    """ The class that will be used to train the student model using the
    distillation knowledge method.

    Reference `Hinton et al. (2015) <https://arxiv.org/abs/1503.02531>`_.

    Args:
        student (keras.Model): the student model
        teacher (keras.Model): the well trained teacher model
    """

    #pylint: disable=abstract-method

    def __init__(self, student, teacher):
        super().__init__()
        self.teacher = teacher
        self.student = student
        self.student_loss_fn = None
        self.distillation_loss_fn = None
        self.alpha = None

    def compile(self,
                optimizer,
                metrics,
                student_loss_fn,
                distillation_loss_fn,
                alpha=0.1):
        """ Configure the distiller.

        Args:
            optimizer (keras.optimizers.Optimizer): Keras optimizer
                for the student weights
            metrics (keras.metrics.Metric): Keras metrics for evaluation
            student_loss_fn (keras.losses.Loss): loss function of difference
                between student predictions and ground-truth
            distillation_loss_fn (keras.losses.Loss): loss function of
                difference between student predictions and teacher predictions
            alpha (float): weight to student_loss_fn and 1-alpha to
                distillation_loss_fn
        """
        #pylint: disable=arguments-differ

        super().compile(optimizer=optimizer, metrics=metrics)
        self.student_loss_fn = student_loss_fn
        self.distillation_loss_fn = distillation_loss_fn
        self.alpha = alpha

    def train_step(self, data):
        # Unpack data
        x, y = data

        # Forward pass of teacher
        teacher_predictions = self.teacher(x, training=False)
        with GradientTape() as tape:
            # Forward pass of student
            student_predictions = self.student(x, training=True)
            # Compute losses
            student_loss = self.student_loss_fn(y, student_predictions)
            distillation_loss = self.distillation_loss_fn(
                teacher_predictions, student_predictions)
            loss = self.alpha * student_loss + (1 -
                                                self.alpha) * distillation_loss

        # Compute gradients
        trainable_vars = self.student.trainable_variables
        gradients = tape.gradient(loss, trainable_vars)

        # Update weights
        self.optimizer.apply_gradients(zip(gradients, trainable_vars))

        # Update the metrics configured in `compile()`.
        self.compiled_metrics.update_state(y, student_predictions)

        # Return a dict of performance
        results = {m.name: m.result() for m in self.metrics}
        results.update({
            "student_loss": student_loss,
            "distillation_loss": distillation_loss
        })
        return results

    def test_step(self, data):
        # Unpack the data
        x, y = data

        # Compute predictions
        y_prediction = self.student(x, training=False)

        # Calculate the loss
        student_loss = self.student_loss_fn(y, y_prediction)

        # Update the metrics.
        self.compiled_metrics.update_state(y, y_prediction)

        # Return a dict of performance
        results = {m.name: m.result() for m in self.metrics}
        results.update({"student_loss": student_loss})
        return results


class KLDistillationLoss(KLDivergence):
    """
    The `KLDistillationLoss` is a simple wrapper around the KLDivergence loss
    that accepts raw predictions instead of probability distributions.

    Before invoking the KLDivergence loss, it converts the inputs predictions to
    probabilities by dividing them by a constant 'temperature' and applies a
    softmax.

    Args:
        temperature (float): temperature for softening probability
            distributions. Larger temperature gives softer distributions.
    """

    def __init__(self, temperature=3):
        self.temperature = temperature
        super().__init__()

    def __call__(self, y_true, y_pred, sample_weight=None):
        return super().__call__(
            tf.nn.softmax(y_true / self.temperature, axis=1),
            tf.nn.softmax(y_pred / self.temperature, axis=1))



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Convtiny model definition for DVS gesture classification.
"""

from keras.utils.data_utils import get_file
from keras import Model
from keras.layers import Input, Reshape, Dropout, Activation

from cnn2snn import load_quantized_model, quantize

from ..layer_blocks import conv_block, separable_conv_block, dense_block

# Locally fixed config options
# The number of neurons in the penultimate dense layer
# This layer has binary output spikes, and could be a bottleneck
# if care isn't taken to ensure enough info capacity
NUM_SPIKING_NEURONS = 256

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/convtiny/'


def convtiny_dvs_gesture(input_shape=(64, 64, 6),
                         classes=None,
                         weight_quantization=0,
                         activ_quantization=0):
    """Instantiates a CNN for the "IBM DVS Gesture" example.

    Args:
        input_shape (tuple): input shape tuple of the model
        classes (int): number of classes to classify images into.
        weight_quantization (int): sets all weights in the model to have
            a particular quantization bitwidth except for the weights in the
            first layer.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization: sets all activations in the model to have a
            particular activation quantization bitwidth.

            * '0' implements floating point 32-bit activations.
            * '1' through '8' implements n-bit weights where n is from 1-8 bits.

    Returns:
        keras.Model: a Keras convolutional model for DVS Gesture.
    """

    img_input = Input(input_shape)

    x = conv_block(img_input,
                   filters=16,
                   kernel_size=(3, 3),
                   name='conv_01',
                   use_bias=False,
                   add_batchnorm=True,
                   padding='same',
                   pooling='max',
                   pool_size=(2, 2),
                   add_activation=True,
                   strides=(1, 1))

    x = conv_block(x,
                   filters=16,
                   kernel_size=(3, 3),
                   name='conv_02',
                   use_bias=False,
                   add_batchnorm=True,
                   padding='same',
                   pooling='max',
                   pool_size=(2, 2),
                   add_activation=True,
                   strides=(1, 1))

    x = conv_block(x,
                   filters=32,
                   kernel_size=(3, 3),
                   name='conv_03',
                   use_bias=False,
                   add_batchnorm=True,
                   padding='same',
                   pooling='max',
                   pool_size=(2, 2),
                   add_activation=True,
                   strides=(1, 1))

    x = conv_block(x,
                   filters=64,
                   kernel_size=(3, 3),
                   name='conv_04',
                   use_bias=False,
                   add_batchnorm=True,
                   padding='same',
                   pooling='max',
                   pool_size=(2, 2),
                   add_activation=True,
                   strides=(1, 1))

    x = conv_block(x,
                   filters=128,
                   kernel_size=(3, 3),
                   name='conv_05',
                   use_bias=False,
                   add_batchnorm=True,
                   padding='same',
                   pooling='global_avg',
                   pool_size=(2, 2),
                   add_activation=True,
                   strides=(1, 1))

    bm_outshape = (1, 1, 128)

    x = Reshape(bm_outshape, name='reshape_1')(x)
    x = Dropout(1e-3, name='dropout')(x)

    x = separable_conv_block(x,
                             filters=NUM_SPIKING_NEURONS,
                             kernel_size=(3, 3),
                             use_bias=False,
                             padding='same',
                             name='spiking_layer',
                             add_batchnorm=True,
                             add_activation=True,
                             pooling=None)

    x = dense_block(x,
                    units=classes,
                    add_batchnorm=False,
                    add_activation=False,
                    use_bias=False)
    act_function = 'softmax' if classes > 1 else 'sigmoid'
    x = Activation(act_function, name=f'act_{act_function}')(x)
    x = Reshape((classes,), name='reshape_2')(x)

    model = Model(inputs=img_input, outputs=x, name='dvs_network')

    if ((weight_quantization != 0) or (activ_quantization != 0)):
        return quantize(model, weight_quantization, activ_quantization)
    return model


def convtiny_gesture_pretrained():
    """ Helper method to retrieve a `convtiny_dvs_gesture` model that was
    trained on IBM DVS Gesture dataset.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'convtiny_dvs_gesture_iq2_wq2_aq4.h5'
    file_hash = '73ca8636954254d85a135b4750c7fdf6a5c5291dd0275139cfe09b11a1cbc20e'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
PointNet++ model definition for ModelNet40 classification.
"""

from keras import layers, Model, regularizers
from keras.utils.data_utils import get_file

from cnn2snn import quantize, load_quantized_model

from .pointnet_utils import get_reshape_factor
from ..layer_blocks import conv_block, dense_block

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/pointnet_plus/'


def pointnet_plus_modelnet40(selected_points=128,
                             features=3,
                             knn_points=64,
                             classes=40,
                             alpha=1.0,
                             weight_quantization=0,
                             activ_quantization=0):
    """ Instantiates a PointNet++ model for the ModelNet40 classification.

    This example implements the point cloud deep learning paper
    `PointNet (Qi et al., 2017) <https://arxiv.org/abs/1612.00593>`_. For a
    detailed introduction on PointNet see `this blog post
    <https://medium.com/@luis_gonzales/an-in-depth-look-at-pointnet-111d7efdaa1a>`_.

    PointNet++ is conceived as a repeated series of operations: sampling and
    grouping of points, followed by the trainable convnet itself. Those
    operations are then repeated at increased scale.
    Each of the selected points is taken as the centroid of the K-nearest
    neighbours. This defines a localized group.

    Args:
        selected_points (int, optional): the number of points to process per
            sample. Default is 128.
        features (int, optional): the number of features. Expected values are
            1 or 3. Default is 3.
        knn_points (int, optional): the number of points to include in each
            localised group. Must be a power of 2, and ideally an integer square
            (so 64, or 16 for a deliberately small network, or 256 for large).
            Default is 64.
        classes (int, optional): the number of classes for the classifier.
            Default is 40.
        alpha (float, optional): network filters multiplier. Default is 1.0.
        weight_quantization (int, optional): sets all weights in the model to
            have a particular quantization bitwidth except for the weights in
            the first layer. Defaults to 0.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization (int, optional): sets all activations in the model to
            have a particular activation quantization bitwidth. Defaults to 0.

            * '0' implements floating point 32-bit activations.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.

    Returns:
        keras.Model: a quantized Keras model for PointNet++/ModelNet40.
    """
    # Adapt input shape with preprocessing
    reshape_factor = get_reshape_factor(knn_points)
    input_shape = (selected_points * reshape_factor,
                   knn_points // reshape_factor, features)

    inputs = layers.Input(shape=input_shape)
    base_filter_num = round(32 * alpha)
    reg = regularizers.l1_l2(1e-7, 1e-7)

    # Block 1
    x = conv_block(inputs,
                   filters=base_filter_num,
                   name='block1_conv1',
                   kernel_size=(3, 3),
                   padding='same',
                   add_batchnorm=True,
                   add_activation=False)
    x = layers.ReLU(max_value=6, activity_regularizer=reg)(x)

    x = conv_block(x,
                   filters=base_filter_num,
                   name='bloc1_conv2',
                   kernel_size=(1, 1),
                   padding='same',
                   add_batchnorm=True,
                   add_activation=False)
    x = layers.ReLU(max_value=6, activity_regularizer=reg)(x)
    x = layers.MaxPool2D(padding='same')(x)

    # Block 2
    x = conv_block(x,
                   filters=base_filter_num * 2,
                   name='block2_conv1',
                   kernel_size=(1, 1),
                   padding='same',
                   add_batchnorm=True,
                   add_activation=False)
    x = layers.ReLU(max_value=6, activity_regularizer=reg)(x)
    x = conv_block(x,
                   filters=base_filter_num * 2,
                   name='block2_conv2',
                   kernel_size=(1, 1),
                   padding='same',
                   add_batchnorm=True,
                   add_activation=False)
    x = layers.ReLU(max_value=6, activity_regularizer=reg)(x)
    if knn_points >= 8:
        x = layers.MaxPool2D(padding='same')(x)

    # Block 3
    x = conv_block(x,
                   filters=base_filter_num * 4,
                   name='block3_conv1',
                   kernel_size=(1, 1),
                   padding='same',
                   add_batchnorm=True,
                   add_activation=False)
    x = layers.ReLU(max_value=6, activity_regularizer=reg)(x)
    x = conv_block(x,
                   filters=base_filter_num * 4,
                   name='block3_conv2',
                   kernel_size=(1, 1),
                   padding='same',
                   add_batchnorm=True,
                   add_activation=False)
    x = layers.ReLU(max_value=6, activity_regularizer=reg)(x)
    if knn_points >= 32:
        x = layers.MaxPool2D(padding='same')(x)

    # Block 4
    x = conv_block(x,
                   filters=base_filter_num * 8,
                   name='block4_conv1',
                   kernel_size=(1, 1),
                   padding='same',
                   add_batchnorm=True,
                   add_activation=False)
    x = layers.ReLU(max_value=6, activity_regularizer=reg)(x)
    if knn_points >= 128:
        x = layers.MaxPool2D(padding='same')(x)

    # Block 5
    x = conv_block(x,
                   filters=base_filter_num * 16,
                   name='block5_conv1',
                   kernel_size=(1, 1),
                   pooling='global_avg',
                   padding='same',
                   add_batchnorm=True)

    # Block 6
    x = dense_block(x,
                    units=base_filter_num * 16,
                    name='fc1',
                    add_batchnorm=True)
    x = dense_block(x,
                    units=base_filter_num * 8,
                    name='fc2',
                    add_batchnorm=True)
    x = layers.Dense(classes, activation=None)(x)
    act_function = 'softmax' if classes > 1 else 'sigmoid'
    outputs = layers.Activation(act_function, name=f'act_{act_function}')(x)

    model = Model(inputs=inputs, outputs=outputs, name="pointnet_plus")

    if ((weight_quantization != 0) or (activ_quantization != 0)):
        # Converts a standard sequential Keras model to a CNN2SNN Keras
        # quantized model, compatible for Akida conversion.
        model = quantize(model=model,
                         weight_quantization=weight_quantization,
                         activ_quantization=activ_quantization,
                         input_weight_quantization=8)

    return model


def pointnet_plus_modelnet40_pretrained():
    """
    Helper method to retrieve a `pointnet_plus` model that was trained on
    ModelNet40 dataset.

    Returns:
        keras.Model: a Keras Model instance.
    """
    model_name = 'pointnet_plus_modelnet40_iq8_wq4_aq4.h5'
    file_hash = '464f9cccc868c46eb4ba16b722a2e7da86678dd8f50ec6bd988b965614022fc9'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')

    return load_quantized_model(model_path)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
This module defines a custom loss function for YOLO training.
"""

import numpy as np
import tensorflow as tf

from keras import backend as K

EPSILON = 1e-7


class YoloLoss():
    """ Computes YOLO loss from a model raw output.

        Custom loss calculation to be given to a Keras model "compile" method.
        YOLO loss uses sum-squared error and is composed of a localisation loss
        (using bounding box coordinates), a confidence loss (based on objects
        IoU) and a classification loss. Full details on the loss computation in
        the original paper https://arxiv.org/pdf/1506.02640.pdf.

        Args:
            anchors (list): anchors associated to the model.
            grid_size (tuple): yolo grid size.
            batch_size (int): the batch size.
            lambda_coord (float, optional): constant used to weight the loss
                of coordinate predictions.
            lambda_noobj (float, optional): constant used to weight the loss
                of empty boxes.
            lambda_obj (float, optional): constant used to weight the loss
                of non-empty boxes.
            lambda_class (float, optional): constant used to weight the
                classification loss.
            iou_filter (float, optional): IoU threshold used in the object loss.
    """

    def __init__(self,
                 anchors,
                 grid_size,
                 batch_size,
                 lambda_coord=1.0,
                 lambda_noobj=1.0,
                 lambda_obj=5.0,
                 lambda_class=1.0,
                 iou_filter=0.6):

        self.__name__ = 'yolo_loss'
        self.iou_filter = iou_filter
        self.readjust_obj_score = False

        self.lambda_coord = lambda_coord
        self.lambda_noobj = lambda_noobj
        self.lambda_obj = lambda_obj
        self.lambda_class = lambda_class

        self.batch_size = batch_size
        self.grid_size = grid_size
        self.nb_anchors = len(anchors)
        self.anchors = np.reshape(anchors, [1, 1, 1, self.nb_anchors, 2])

        self.c_grid = self._generate_yolo_grid(self.batch_size, self.grid_size,
                                               self.nb_anchors)

    @staticmethod
    def _generate_yolo_grid(batch_size, grid_size, nb_box):
        cell_x = tf.cast(
            tf.reshape(tf.tile(tf.range(grid_size[0]), [grid_size[1]]),
                       (1, grid_size[1], grid_size[0], 1, 1)), tf.float32)
        cell_y = tf.cast(
            tf.reshape(tf.tile(tf.range(grid_size[1]), [grid_size[0]]),
                       (1, grid_size[0], grid_size[1], 1, 1)), tf.float32)
        cell_y = tf.transpose(cell_y, (0, 2, 1, 3, 4))

        cell_grid = tf.tile(tf.concat([cell_x, cell_y], -1),
                            [batch_size, 1, 1, nb_box, 1])
        return cell_grid

    def _transform_netout(self, y_pred_raw):

        y_pred_xy = K.sigmoid(y_pred_raw[..., :2]) + self.c_grid
        y_pred_wh = K.exp(y_pred_raw[..., 2:4]) * self.anchors
        y_pred_conf = K.sigmoid(y_pred_raw[..., 4:5])
        y_pred_class = y_pred_raw[..., 5:]

        return K.concatenate([y_pred_xy, y_pred_wh, y_pred_conf, y_pred_class],
                             axis=-1)

    def coord_loss(self, y_true, y_pred):
        """ Computes the localisation loss.

        Args:
            y_true (tf.Tensor): tensor of true labels
            y_pred (tf.Tensor): tensor of predicted labels

        Returns:
            float: the computed loss
        """

        b_xy_pred = y_pred[..., :2]
        b_wh_pred = y_pred[..., 2:4]

        b_xy = y_true[..., 0:2]
        b_wh = y_true[..., 2:4]

        indicator_coord = K.expand_dims(y_true[..., 4],
                                        axis=-1) * self.lambda_coord

        loss_xy = K.sum(K.square(b_xy - b_xy_pred) * indicator_coord)
        loss_wh = K.sum(K.square(b_wh - b_wh_pred) * indicator_coord)

        return (loss_wh + loss_xy) / 2

    def obj_loss(self, y_true, y_pred):
        """ Computes the confidence loss.

        Args:
            y_true (tf.Tensor): tensor of true labels
            y_pred (tf.Tensor): tensor of predicted labels

        Returns:
            float: the computed loss
        """

        b_o = self._calculate_ious(y_true,
                                   y_pred,
                                   use_iou=self.readjust_obj_score)
        b_o_pred = y_pred[..., 4]

        num_true_labels = self.grid_size[0] * self.grid_size[1] * self.nb_anchors
        y_true_p = K.reshape(y_true[..., :4],
                             shape=(self.batch_size, 1, 1, 1, num_true_labels,
                                    4))
        iou_scores_buff = self._calculate_ious(y_true_p,
                                               K.expand_dims(y_pred, axis=4))
        best_ious = K.max(iou_scores_buff, axis=4)

        indicator_noobj = K.cast(best_ious < self.iou_filter, np.float32) * (
            1 - y_true[..., 4]) * self.lambda_noobj
        indicator_obj = y_true[..., 4] * self.lambda_obj
        indicator_o = indicator_obj + indicator_noobj

        loss_obj = K.sum(K.square(b_o - b_o_pred) * indicator_o)
        return loss_obj / 2

    def class_loss(self, y_true, y_pred):
        """ Computes the classification loss.

        Args:
            y_true (tf.Tensor): tensor of true labels
            y_pred (tf.Tensor): tensor of predicted labels

        Returns:
            float: the computed loss
        """

        p_c_pred = K.softmax(y_pred[..., 5:])
        p_c = y_true[..., 5:]
        loss_class_arg = K.sum(K.square(p_c - p_c_pred), axis=-1)

        indicator_class = y_true[..., 4] * self.lambda_class

        loss_class = K.sum(loss_class_arg * indicator_class)

        return loss_class

    def __call__(self, y_true, y_pred_raw):

        y_pred = self._transform_netout(y_pred_raw)

        total_coord_loss = self.coord_loss(y_true, y_pred)
        total_obj_loss = self.obj_loss(y_true, y_pred)
        total_class_loss = self.class_loss(y_true, y_pred)

        loss = total_coord_loss + total_obj_loss + total_class_loss

        return loss

    @staticmethod
    def _calculate_ious(a1, a2, use_iou=True):

        if not use_iou:
            return a1[..., 4]

        def process_boxes(a):
            # Align x-w, y-h
            a_xy = a[..., 0:2]
            a_wh = a[..., 2:4]

            a_wh_half = a_wh / 2.
            # Get x_min, y_min
            a_mins = a_xy - a_wh_half
            # Get x_max, y_max
            a_maxes = a_xy + a_wh_half

            return a_mins, a_maxes, a_wh

        # Process two sets
        a2_mins, a2_maxes, a2_wh = process_boxes(a2)
        a1_mins, a1_maxes, a1_wh = process_boxes(a1)

        # Intersection as min(Upper1, Upper2) - max(Lower1, Lower2)
        intersect_mins = K.maximum(a2_mins, a1_mins)
        intersect_maxes = K.minimum(a2_maxes, a1_maxes)

        # Getting the intersections in the xy (aka the width, height
        # intersection)
        intersect_wh = K.maximum(intersect_maxes - intersect_mins, 0.)

        # Multiply to get intersecting area
        intersect_areas = intersect_wh[..., 0] * intersect_wh[..., 1]

        # Values for the single sets
        true_areas = a1_wh[..., 0] * a1_wh[..., 1]
        pred_areas = a2_wh[..., 0] * a2_wh[..., 1]

        # Compute union for the IoU
        union_areas = pred_areas + true_areas - intersect_areas
        return intersect_areas / union_areas



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Preprocessing tools for CSE-CIC-IDS-2018 data handling.
"""

import os
import pickle
import tensorflow as tf

from keras.utils.data_utils import get_file


def load_labels_mapping():
    """ Loads labels mapping from Brainchip data server.

    Returns:
        dict: the dictionnary of labels mapping
    """
    labels_mapping_path = get_file(
        "labels_mapping.pkl",
        "http://data.brainchip.com/dataset-mirror/cse_cic_ids_2018/labels_mapping.pkl",
        cache_subdir='datasets/cse_2018/')

    with open(labels_mapping_path, 'rb') as handle:
        labels_mapping = pickle.load(handle)
    return labels_mapping


def load_columns_type():
    """ Loads columns types from Brainchip data server.

    Returns:
        list, dict: the list of numerical columns and the dictionary of
        categorical columns with their vocabulary
    """
    columns_types_path = get_file(
        "columns_types.pkl",
        "http://data.brainchip.com/dataset-mirror/cse_cic_ids_2018/columns_types.pkl",
        cache_subdir='datasets/cse_2018/')

    with open(columns_types_path, 'rb') as handle:
        numerical_columns, categorical_columns = pickle.load(handle)
    return numerical_columns, categorical_columns


def load_data(batch_size=65536, load_train=True, load_test=True):
    """ Loads the dataset from Brainchip data server.

    Args:
        batch_size (int): number of records to combine in a single batch
        load_train (bool): flag to enable train set loading
        load_test (bool): flag to enable test set loading

    Returns:
        tf.dataset, tf.dataset: the train and test sets

    """
    LABEL_COLUMN = 'label'

    x_train = None
    x_test = None

    if load_train:
        train_file_path = get_file(
            "train_set.zip",
            "http://data.brainchip.com/dataset-mirror/cse_cic_ids_2018/train_set.zip",
            cache_subdir='datasets/cse_2018/',
            extract=True)
        train_file_path = os.path.join(os.path.dirname(train_file_path),
                                       "train_set.csv")
        x_train = tf.data.experimental.make_csv_dataset(train_file_path,
                                                        batch_size=batch_size,
                                                        label_name=LABEL_COLUMN,
                                                        num_epochs=1,
                                                        ignore_errors=True)

    if load_test:
        test_file_path = get_file(
            "test_set.zip",
            "http://data.brainchip.com/dataset-mirror/cse_cic_ids_2018/test_set.zip",
            cache_subdir='datasets/cse_2018/',
            extract=True)
        test_file_path = os.path.join(os.path.dirname(test_file_path),
                                      "test_set.csv")
        x_test = tf.data.experimental.make_csv_dataset(test_file_path,
                                                       batch_size=batch_size,
                                                       label_name=LABEL_COLUMN,
                                                       num_epochs=1,
                                                       ignore_errors=True)

    return x_train, x_test



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
This model is an adaptation of the `mobilenet_imagenet` model for edge
applications. It is based on MobileNetBC with top layers replaced by a quantized
spike extractor and a classification layer.
"""

from keras import Model
from keras.utils.data_utils import get_file
from keras.layers import Reshape, Activation

from cnn2snn import load_quantized_model, quantize_layer

from ..layer_blocks import separable_conv_block, dense_block

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/mobilenet_edge/'


def mobilenet_edge_imagenet(base_model, classes):
    """Instantiates a MobileNet-edge architecture.

    Args:
        base_model (str/keras.Model): a mobilenet_imagenet quantized model.
        classes (int) : the number of classes for the edge classifier.

    Returns:
        keras.Model: a Keras Model instance.
    """
    if isinstance(base_model, str):
        base_model = load_quantized_model(base_model)

    try:
        # Identify the last separable, which is the base model classifier
        base_classifier = base_model.get_layer("separable_14")
        # remember the classifier weight bitwidth
        wq = base_classifier.quantizer.bitwidth
    except Exception as e:
        raise ValueError("The base model is not a quantized \
                          Mobilenet/Imagenet model") from e

    # Recreate a model with all layers up to the classifier
    x = base_classifier.input
    # Add the new end layer with kernel_size (3, 3) instead of (1,1) for
    # hardware compatibility reasons
    x = separable_conv_block(x,
                             filters=2048,
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             name='spike_generator',
                             add_activation=True)

    # Then add the Akida edge learning layer that will be dropped after
    x = dense_block(x,
                    classes,
                    name="classification_layer",
                    add_activation=False,
                    add_batchnorm=False,
                    use_bias=False)
    act_function = 'softmax' if classes > 1 else 'sigmoid'
    x = Activation(act_function, name=f'act_{act_function}')(x)
    x = Reshape((classes,), name='reshape_2')(x)

    # Create model
    model = Model(inputs=base_model.input,
                  outputs=x,
                  name=f"{base_model.name}_edge")

    # Quantize edge layers
    model = quantize_layer(model, 'spike_generator', wq)
    model = quantize_layer(model, 'spike_generator_relu', 1)
    # NOTE: quantization set to 2 here, to be as close as
    # possible to the Akida native layer that will replace this one,
    # with binary weights.
    model = quantize_layer(model, 'classification_layer', 2)

    return model


def mobilenet_edge_imagenet_pretrained():
    """ Helper method to retrieve a `mobilenet_edge_imagenet` model that was
    trained on ImageNet dataset.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'mobilenet_imagenet_224_alpha_50_edge_iq8_wq4_aq4.h5'
    file_hash = 'b759ed6aa5b47923e5cee03f756e63d7a5865e49f34749e55aff242d83b9ee75'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
ModelNet40 model training script.
"""

import argparse
import keras
import numpy as np

from keras.callbacks import LearningRateScheduler

from cnn2snn import load_quantized_model, convert

from .preprocessing import get_modelnet, get_modelnet_from_file

from ..training import get_training_parser, compile_model


def get_data(batch_size, selected_points, knn_points):
    """ Loads data.

    Args:
        batch_size (int): the batch size
        selected_points (int): num points to process per sample.
        knn_points (int): number of points to include in each localised group.
            Must be a power of 2, and ideally an integer square (so 64, or 16
            for a deliberately small network, or 256 for large).

    Returns:
        tf.Dataset, tf.Dataset: train and test point with data augmentation
    """
    # This is the number of points sampled from each input file,
    # when parsing the data, and done only once (and saved as a .npz file)
    # for efficiency.
    train_points, train_labels, test_points, test_labels, _ = \
        get_modelnet_from_file(num_points=1024)

    train_dataset, test_dataset = get_modelnet(train_points,
                                               train_labels,
                                               test_points,
                                               test_labels,
                                               batch_size=batch_size,
                                               selected_points=selected_points,
                                               knn_points=knn_points)

    return train_dataset, test_dataset


def train_model(model, train_dataset, test_dataset, epochs):
    """ Trains the model.

    Args:
        model (keras.Model): the model to train.
        train_dataset (tf.dataset): train data.
        test_dataset (tf.dataset): test data.
        epochs (int): the number of epochs.
    """
    # Learning rate scheduler
    callbacks = []
    max_lr = 0.001
    min_lr = max_lr * 0.01
    lr_decay = (min_lr / max_lr)**(1. / epochs)
    lr_scheduler = LearningRateScheduler(lambda e: max_lr * lr_decay**e)
    callbacks.append(lr_scheduler)

    history = model.fit(train_dataset,
                        validation_data=test_dataset,
                        epochs=epochs,
                        callbacks=callbacks)
    print(history.history)


def evaluate_model(model, test_dataset):
    """ Evaluates model performances.

    Args:
        model (keras.Model or akida.Model): the model to evaluate.
        test_dataset (tf.dataset): evaluation dataset.
    """
    labels = np.concatenate([labels for x, labels in test_dataset], axis=0)

    if isinstance(model, keras.Model):
        preds = np.squeeze(np.argmax(model.predict(test_dataset), axis=1))
    else:
        pots_akida = np.empty(shape=[0, model.output_shape[-1]],
                              dtype=np.float32)

        for batch, _ in test_dataset:
            # Rescale inputs to Akida
            pots = (batch.numpy() * 127 + 127).astype('uint8')
            pots_batch = model.evaluate(pots)
            pots_akida = np.concatenate((pots_akida, pots_batch.squeeze()))

        preds = np.squeeze(np.argmax(pots_akida, 1))

    acc = np.sum(preds == labels) / labels.shape[0]
    print(f"Accuracy: {acc}")


def main():
    """ Entry point for script and CLI usage.
    """
    global_parser = argparse.ArgumentParser(add_help=False)
    global_parser.add_argument(
        "-p",
        "--selected_points",
        type=int,
        default=128,
        help="The number of points to process per sample. \
                        (default 128)")
    global_parser.add_argument("-n",
                               "--knn_points",
                               type=int,
                               default=64,
                               help="The number of points to include in each \
                        localised group. (default 64)")

    parser = get_training_parser(batch_size=16, global_parser=global_parser)[0]
    args = parser.parse_args()

    # Load the source model
    model = load_quantized_model(args.model)

    # Compile model
    compile_model(model, loss="sparse_categorical_crossentropy")

    # Load data
    train_dataset, test_dataset = get_data(args.batch_size,
                                           args.selected_points,
                                           args.knn_points)

    # Train model
    if args.action == "train":
        train_model(model, train_dataset, test_dataset, args.epochs)

        # Save model in Keras format (h5)
        if args.savemodel is not None:
            model.save(args.savemodel, include_optimizer=False)
            print(f"Trained model saved as {args.savemodel}")

    elif args.action == "eval":
        # Evaluate model accuracy
        if args.akida:
            model = convert(model, input_scaling=(127, 127))
        evaluate_model(model, test_dataset)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
KWS model training script.
"""

import os
import pickle

from keras.utils.np_utils import to_categorical
from keras.utils.data_utils import get_file

from cnn2snn import load_quantized_model, quantize_layer, convert

from ..cyclic_lr import CyclicLR
from ..training import get_training_parser, compile_model, evaluate_model


def get_data():
    """ Loads KWS data.

    Returns:
        tuple: train data, train labels, validation data and validation labels
    """
    # Load pre-processed dataset
    fname = get_file(
        fname='kws_preprocessed_all_words_except_backward_follow_forward.pkl',
        origin=
        'http://data.brainchip.com/dataset-mirror/kws/kws_preprocessed_all_words_except_backward_follow_forward.pkl',
        cache_subdir=os.path.join('datasets', 'kws'))

    print('Loading pre-processed dataset...')
    with open(fname, 'rb') as f:
        [x_train, y_train, x_val, y_val, _, _, _, _] = pickle.load(f)

    return x_train.astype('float32'), y_train, x_val.astype('float32'), y_val


def train_model(model, x_train, y_train, x_val, y_val, batch_size, epochs):
    """ Trains the model.

    Args:
        model (keras.Model): the model to train
        x_train (numpy.ndarray): train data
        y_train (numpy.ndarray): train labels
        x_val (numpy.ndarray): validation data
        y_val (numpy.ndarray): validation labels
        batch_size (int): the batch size
        epochs (int): the number of epochs
    """
    # Warn user if number of epochs is not a multiple of 8
    if epochs % 8:
        print("WARNING: for better performance, the number of epochs "
              f" must be a multiple of 8; got 'epochs' = {epochs}")

    # Training parameters (cyclical learning rate)
    scaler = 4
    base_lr = 5e-6
    max_lr = 2e-3

    # Cyclical learning rate
    callbacks = []
    clr = CyclicLR(base_lr=base_lr,
                   max_lr=max_lr,
                   step_size=scaler * x_train.shape[0] / batch_size,
                   mode='triangular')
    callbacks.append(clr)

    history = model.fit(x_train,
                        to_categorical(y_train),
                        batch_size=batch_size,
                        epochs=epochs,
                        verbose=1,
                        validation_data=(x_val, to_categorical(y_val)),
                        callbacks=callbacks)
    print(history.history)


def main():
    """ Entry point for script and CLI usage.
    """
    parsers = get_training_parser(batch_size=100, global_batch_size=False)

    train_parser = parsers[1]
    train_parser.add_argument("-laq",
                              "--last_activ_quantization",
                              type=int,
                              default=None,
                              help="The last layer activation quantization")

    args = parsers[0].parse_args()

    # Load the source model
    model = load_quantized_model(args.model)

    # If specified, change the last layer activation bitwidth
    if "last_activ_quantization" in args and args.last_activ_quantization:
        model = quantize_layer(model, 'separable_4_relu',
                               args.last_activ_quantization)

    # Compile model
    compile_model(model, learning_rate=2e-3)

    # Load data
    x_train, y_train, x_val, y_val = get_data()

    # Train model
    if args.action == "train":
        train_model(model, x_train, y_train, x_val, y_val, args.batch_size,
                    args.epochs)

        # Save model in Keras format (h5)
        if args.savemodel:
            model.save(args.savemodel, include_optimizer=False)
            print(f"Trained model saved as {args.savemodel}")

    elif args.action == "eval":
        # Evaluate model accuracy
        if args.akida:
            model_ak = convert(model)
            preds = model_ak.predict(x_val.astype('uint8'))
            accuracy = (y_val == preds).mean()
            print(f"Accuracy: {accuracy}")
        else:
            evaluate_model(model, x_val, y=to_categorical(y_val))


if __name__ == "__main__":
    main()



# ******************************************************************************
# MIT License
#
# Copyright (c) 2017 Ngoc Anh Huynh
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ******************************************************************************
"""
This module provides a method to generate YOLO anchors from dataset annotations.
"""

import random
import numpy as np


def _iou(ann, centroids):
    w, h = ann
    similarities = []

    for centroid in centroids:
        c_w, c_h = centroid

        if c_w >= w and c_h >= h:
            similarity = w * h / (c_w * c_h)
        elif c_w >= w and c_h <= h:
            similarity = w * c_h / (w * h + (c_w - w) * c_h)
        elif c_w <= w and c_h >= h:
            similarity = c_w * h / (w * h + c_w * (c_h - h))
        else:  #means both w,h are bigger than c_w and c_h respectively
            similarity = (c_w * c_h) / (w * h)
        similarities.append(similarity)  # will become (k,) shape

    return np.array(similarities)


def _avg_iou(anns, centroids):
    n, _ = anns.shape
    s = 0.

    for i in range(anns.shape[0]):
        s += max(_iou(anns[i], centroids))

    return s / n


def _run_kmeans(ann_dims, anchor_num):
    ann_num = ann_dims.shape[0]
    prev_assignments = np.ones(ann_num) * (-1)
    iteration = 0

    indices = [random.randrange(ann_dims.shape[0]) for _ in range(anchor_num)]
    centroids = ann_dims[indices]
    anchor_dim = ann_dims.shape[1]

    while True:
        distances = []
        iteration += 1
        for i in range(ann_num):
            d = 1 - _iou(ann_dims[i], centroids)
            distances.append(d)
        distances = np.array(distances)

        # assign samples to centroids
        assignments = np.argmin(distances, axis=1)

        if (assignments == prev_assignments).all():
            return centroids

        #calculate new centroids
        centroid_sums = np.zeros((anchor_num, anchor_dim), np.float)
        for i in range(ann_num):
            centroid_sums[assignments[i]] += ann_dims[i]
        for j in range(anchor_num):
            centroids[j] = centroid_sums[j] / (np.sum(assignments == j) + 1e-6)

        prev_assignments = assignments.copy()


def generate_anchors(annotations_data, num_anchors=5, grid_size=(7, 7)):
    """ Creates anchors by clustering dimensions of the ground truth boxes
    from the training dataset.

    Args:
        annotations_data (dict): dictionnary of preprocessed VOC data
        num_anchors (int, optional): number of anchors
        grid_size (tuple, optional): size of the YOLO grid

    Returns:
        list: the computed anchors
    """
    annotation_dims = []

    # run k_mean to find the anchors
    for item in annotations_data:
        cell_w = item['image_shape'][0] / grid_size[0]
        cell_h = item['image_shape'][1] / grid_size[1]

        for box in item['boxes']:
            relative_w = float(box['x2'] - box['x1']) / cell_w
            relative_h = float(box['y2'] - box['y1']) / cell_h
            annotation_dims.append(tuple(map(float, (relative_w, relative_h))))

    annotation_dims = np.array(annotation_dims)
    centroids = _run_kmeans(annotation_dims, num_anchors)
    print('\nAverage IOU for', num_anchors, 'anchors:',
          '%0.2f' % _avg_iou(annotation_dims, centroids))
    anchors = np.sort(np.round(centroids, 5), 0).tolist()
    print('Anchors: ', anchors)
    return anchors



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
TSE and MLP model definitions.
"""

import numpy as np
import tensorflow as tf

from keras.layers import Input, Activation, Reshape, Flatten
from keras.feature_column.dense_features_v2 import DenseFeatures

from ..layer_blocks import dense_block


def create_tse(numerical_columns, categorical_columns, units):
    """
    Create a Trainable Spike Encoder (TSE) network.

    Since Akida cannot handle a flat Dense as input, a reshape layer is inserted
    as the last layer the TSE sub-model and the first layer of the MLP part is a
    Flatten layer.

    Note that the following rule is applied to choose the embedding size of
    categorical columns:

    'Inspired by Howard & Gugger (2020), a discrete feature with k unique
    categories observed in the training data is processed by an embedding layer
    of size: 1.6 * k^0.56 (up to threshold of size 100).'

    See:
        - Howard, J., & Gugger, S. (2020). Deep Learning for Coders with fastai
          and PyTorch: AI Applications Without a PhD. 1st Edition (in particular
          Chapter 9: `09_tabular.ipynb <https://github.com/fastai/fastbook/blob/master/09_tabular.ipynb>`_
        - Erickson, N., Mueller, J., Shirkov, A., Zhang, H., Larroy, P., Li, M.,
          & Smola, A. (2020). AutoGluon-Tabular: Robust and Accurate AutoML for
          Structured Data. arXiv preprint arXiv:2003.06505.

    Args:
        numerical_columns (list): list of numerical columns names
        categorical_columns (dict): dictionary of categorical column names
            indexing the list of vocabulary for the column
        units (int): : dimensionality of the output space
        input_weight_quantization (int): sets weight quantization in the first
            layer.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization (int): sets all activations in the model to have a
            particular activation quantization bitwidth.

            * '0' implements floating point 32-bit activations.
            * '1' through '8' implements n-bit weights where n is from 1-8 bits.

    Returns:
        dict, tf.Tensor: the dictionary of input layers to build the
        model and the TSE output that can be used as an input for the MLP


    """
    feature_columns = []
    feature_layer_inputs = {}

    for col in numerical_columns:
        feature_layer_inputs[col] = Input(shape=(1,), name=col)
        feature_columns.append(tf.feature_column.numeric_column(col))

    for col, vocab in categorical_columns.items():
        feature_layer_inputs[col] = Input(shape=(1,), name=col, dtype='string')
        cat_col = tf.feature_column.categorical_column_with_vocabulary_list(
            col, vocab, num_oov_buckets=1)
        embedding_size = min(int(np.round(1.6 * len(vocab)**0.56)), 100)
        embedding_col = tf.feature_column.embedding_column(
            cat_col, embedding_size)
        feature_columns.append(embedding_col)

    feature_layer = DenseFeatures(feature_columns)
    feature_layer_outputs = feature_layer(feature_layer_inputs)

    x = dense_block(feature_layer_outputs,
                    units=units,
                    name='dense_to_spike',
                    add_activation=True,
                    add_batchnorm=True)
    x = Reshape(target_shape=(1, 1, units), name='final_embedding_reshape')(x)

    return feature_layer_inputs, x


def create_mlp(inputs, units_list, n_classes):
    """
    Create a Multilayer Perceptron (MLP) network.

    Since Akida cannot handle a flat Dense as input, a reshape layer is inserted
    as the last layer the TSE sub-model and the first layer of the MLP part is a
    Flatten layer.

    Args:
        inputs (:obj:`tf.Tensor`): input tensor of shape (1, 1, N)
        units_list (list): list of units as integers. Units are the
            dimensionality of the output space. The length of the list defines
            the number of hidden layers
        n_classes (int): dimensionality of the last layer output space

    Returns:
        tf.Tensor: the MLP output that will be used to create the model

    """
    x = Flatten()(inputs)

    for i, _ in enumerate(units_list):
        x = dense_block(x,
                        units=units_list[i],
                        name='dense' + str(i + 1),
                        add_activation=True,
                        add_batchnorm=True)
    x = dense_block(x,
                    units=n_classes,
                    name='last_dense',
                    add_activation=False,
                    use_bias=False)
    activation_function = 'softmax' if n_classes > 1 else 'sigmoid'
    x = Activation(activation_function, name='activation')(x)
    return x



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Convtiny model definition for CWRU classification.
"""

from keras import Model
from keras.layers import Input, Softmax, Rescaling, Reshape
from keras.utils.data_utils import get_file

from cnn2snn import load_quantized_model

from ..layer_blocks import conv_block, separable_conv_block

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/convtiny/'


def convtiny_cwru():
    """ Instantiates a CNN for CWRU classification with input shape (32, 32, 1)
    and 10 classes.

    Returns:
        keras.Model: a Keras model for Convtiny/CWRU

    """
    img_input = Input(shape=(32, 32, 1))
    x = Rescaling(1, -127)(img_input)
    x = conv_block(x,
                   filters=32,
                   name='conv_1',
                   kernel_size=(7, 7),
                   padding='same',
                   add_activation=True,
                   pooling='max',
                   pool_size=(2, 2))

    x = conv_block(x,
                   filters=32,
                   name='conv_2',
                   kernel_size=(7, 7),
                   padding='same',
                   add_activation=True,
                   pooling='max',
                   pool_size=(2, 2))

    x = separable_conv_block(x,
                             filters=512,
                             name='separable_1',
                             kernel_size=(3, 3),
                             padding='same',
                             pooling='global_avg',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True)

    x = Reshape((1, 1, int(512)), name='reshape_1')(x)

    x = separable_conv_block(x,
                             filters=10,
                             name='predictions',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=False,
                             add_activation=False)
    x = Reshape((10,), name='reshape_2')(x)
    x = Softmax(name='act_softmax')(x)

    return Model(img_input, x, name='convtiny_cwru_32_10')


def convtiny_cwru_pretrained():
    """
    Helper method to retrieve a `convtiny_cwru` model that was trained on
    CWRU dataset.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'convtiny_cwru_iq8_wq2_aq4.h5'
    file_hash = '15bc18ea85942185d90dbcd35300d86f8751bfd19ff175589d81e05dc2854d40'
    model_path = get_file(fname=model_name,
                          origin=BASE_WEIGHT_PATH + model_name,
                          file_hash=file_hash,
                          cache_subdir='models')
    return load_quantized_model(model_path)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Training script for CIFAR10 models.
"""

import numpy as np

from keras.datasets import cifar10
from keras.callbacks import LearningRateScheduler, EarlyStopping
from keras.utils.np_utils import to_categorical
from keras.preprocessing.image import ImageDataGenerator

from cnn2snn import load_quantized_model, convert

from ..training import get_training_parser, compile_model, evaluate_model


def get_data(target_ak):
    """ Loads CIFAR10 data.

    Args:
        target_ak (bool): whether the target is an Akida model

    Returns:
        tuple: train data, train labels, test data and test labels
    """
    #The data, shuffled and split between train and test sets
    (x_train, y_train), (x_test, y_test) = cifar10.load_data()

    if not target_ak:
        # Convert class vectors to binary class matrices.
        # For hinge loss, we need to rescale them between -1 and 1
        y_train = to_categorical(y_train, 10) * 2 - 1
        y_test = to_categorical(y_test, 10) * 2 - 1

    return x_train.astype('float32'), y_train, x_test.astype('float32'), y_test


def train_model(model, x_train, y_train, x_test, y_test, batch_size, epochs):
    """ Trains the model.

    Args:
        model (keras.Model): the model to train
        x_train (numpy.ndarray): train data
        y_train (numpy.ndarray): train labels
        x_test (numpy.ndarray): test data
        y_test (numpy.ndarray): test labels
        batch_size (int): the batch size
        epochs (int): the number of epochs
    """

    callbacks = []

    # Learning rate: be more aggressive at the beginning, and apply decay
    lr_start = 1e-3
    lr_end = 1e-4
    lr_decay = (lr_end / lr_start)**(1. / epochs)

    lr_scheduler = LearningRateScheduler(lambda e: lr_start * lr_decay**e)
    callbacks.append(lr_scheduler)

    # Use data augmentation
    datagen_args = {}
    # random hz image shift (fraction of width)
    datagen_args['width_shift_range'] = 0.1
    # random vert image shft (fraction of height)
    datagen_args['height_shift_range'] = 0.1
    # Randomly flip images
    datagen_args['horizontal_flip'] = True

    datagen = ImageDataGenerator(**datagen_args)

    training_data = datagen.flow(x_train, y_train, batch_size=batch_size)

    history = model.fit(training_data,
                        steps_per_epoch=len(x_train) / batch_size,
                        epochs=epochs,
                        verbose=1,
                        validation_data=(x_test, y_test),
                        callbacks=callbacks)
    print(history.history)


def tune_model(model, x_train, y_train, x_test, y_test, batch_size, epochs):
    """ Tunes the model.

    Similar to train but with a lower learning rate.

    Args:
        model (keras.Model): the model to train
        x_train (numpy.ndarray): train data
        y_train (numpy.ndarray): train labels
        x_test (numpy.ndarray): test data
        y_test (numpy.ndarray): test labels
        batch_size (int): the batch size
        epochs (int): the number of epochs
    """
    # Early stop when loss has stopped decreasing
    callbacks = []
    es = EarlyStopping(monitor='loss',
                       mode='min',
                       patience=20,
                       restore_best_weights=True)
    callbacks.append(es)

    history = model.fit(x_train,
                        y_train,
                        batch_size=batch_size,
                        epochs=epochs,
                        verbose=1,
                        validation_data=(x_test, y_test),
                        callbacks=callbacks)
    print(history.history)


def main():
    """ Entry point for script and CLI usage.
    """
    parser = get_training_parser(batch_size=500,
                                 tune=True,
                                 global_batch_size=False)[0]
    args = parser.parse_args()

    # Load the source model
    model = load_quantized_model(args.model)

    # Compile model
    learning_rate = 1e-3
    if args.action == "tune":
        learning_rate = 1e-4

    compile_model(model, learning_rate=learning_rate, loss='squared_hinge')

    # Load data
    x_train, y_train, x_test, y_test = get_data("akida" in args and args.akida)

    # Train or tune
    if args.action in ["train", "tune"]:
        if args.action == "train":
            train_model(model, x_train, y_train, x_test, y_test,
                        args.batch_size, args.epochs)
        else:
            tune_model(model, x_train, y_train, x_test, y_test, args.batch_size,
                       args.epochs)

        # Save Model in Keras format (h5)
        if args.savemodel:
            model.save(args.savemodel, include_optimizer=False)
            print(f"Trained model saved as {args.savemodel}")

    elif args.action == "eval":
        # Evaluate model accuracy
        if args.akida:
            model_ak = convert(model)
            preds = model_ak.predict(x_test.astype('uint8'))
            accuracy = (np.squeeze(y_test) == preds).mean()
            print(f"Accuracy: {accuracy}")
        else:
            evaluate_model(model, x_test, y=y_test)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Convtiny/CWRU training script.

CWRU is a standard database used in condition monitoring applications. We have
used the drive end data that are sampled at 48 kHz sampling frequency when load
of 1 hp is applied on the shaft. Normal data collected with 1 hp load have also
been used. There are a total of 10 classes. The classes are:

  - C1 : Ball defect (0.007 inch, load: 1 hp)
  - C2 : Ball defect (0.014 inch, load: 1 hp)
  - C3 : Ball defect (0.021 inch, load: 1 hp)
  - C4 : Inner race fault (0.007 inch, load: 1 hp)
  - C5 : Inner race fault (0.014 inch, load: 1 hp)
  - C6 : Inner race fault (0.021 inch, load: 1 hp)
  - C7 : Normal (load: 1 hp)
  - C8 : Outer race fault (0.007 inch, load: 1 hp, data collected from 6 O'clock
    position)
  - C9 : Outer race fault (0.014 inch, load: 1 hp, 6 O'clock)
  - C10 : Outer race fault (0.021 inch, load: 1 hp, 6 O'clock)

From each category data are collected in segments of length 1024 and resized to
a 2-D matrix of size (32 by 32). There is no overlap between segments. For each
category 460 such segments are taken. Total size of the data thus becomes (4600,
32, 32). Out of this 1000 segments are randomly chosen as test data and rest are
used for training.
"""

import argparse
import numpy as np

from keras.utils.np_utils import to_categorical

from cnn2snn import load_quantized_model, convert

from ..training import get_training_parser, compile_model, evaluate_model


def get_data(data_file, input_size):
    """ Loads and prepares data.

    Args:
        data_file (str): file path to the preprocessed npz data
        input_size (int): size to reshape input data

    Returns:
        np.array: four objects containing training data and labels and test data
            and labels
    """
    # Load data file
    raw_data = np.load(data_file)
    data = raw_data['data']
    labels = raw_data['labels']

    # Transform labels strings to categories
    category_dict = {k: v for v, k in enumerate(np.unique(labels).tolist())}
    labels = np.array([category_dict[label] for label in labels], dtype=np.int8)

    # Rescale data to 8-bit, range [0, 255], the model will rescale inputs so
    # that they are zero centered (i.e [-127, 127])
    absmax = np.maximum(data.max(), -data.min())
    data = np.round(data / absmax * 127) + 127

    # Randomly shuffle data
    np.random.seed(42)
    index = np.random.permutation(len(labels))
    data, labels = data[index], labels[index]

    # Split data: there are only 4600 examples in total in CWRU dataset, thus
    # keeping the test set at 1000 examples which is slightly above 21%.
    splitter = 3600
    train_data = data[:splitter]
    train_labels = labels[:splitter]
    test_data = data[splitter:]
    test_labels = labels[splitter:]

    # Reshape data
    train_data = train_data.reshape(len(train_data), input_size, input_size, 1)
    test_data = test_data.reshape(len(test_data), input_size, input_size, 1)
    train_labels = to_categorical(train_labels)
    test_labels = to_categorical(test_labels)

    # Shuffle data
    index = np.random.permutation(len(train_labels))
    train_data, train_labels = train_data[index], train_labels[index]

    return train_data, train_labels, test_data, test_labels


def train_model(model, train_data, train_labels, batch_size, epochs):
    """ Trains the model.

    Args:
        model (keras.Model): the model to train
        train_data (np.array): train data
        train_labels (np.array): train labels
        batch_size (int): the batch size
        epochs (int): the number of epochs
    """
    history = model.fit(train_data,
                        train_labels,
                        batch_size=batch_size,
                        epochs=epochs)
    print(history.history)


def main():
    """ Script entry point.
    """
    global_parser = argparse.ArgumentParser(add_help=False)
    global_parser.add_argument("-d",
                               "--data",
                               type=str,
                               default='./CWRU_48k_load_1_CNN_data.npz',
                               help="File path to the preprocessed npz data.")

    parser = get_training_parser(batch_size=128, global_parser=global_parser)[0]
    args = parser.parse_args()

    # Load the source model
    model = load_quantized_model(args.model)

    # Compile model
    compile_model(model)

    # Load data
    train_data, train_labels, test_data, test_labels = get_data(
        args.data, model.input_shape[1])

    # Train model
    if args.action == "train":
        train_model(model, train_data, train_labels, args.batch_size,
                    args.epochs)

        # Save model in Keras format (h5)
        if args.savemodel:
            model.save(args.savemodel, include_optimizer=False)
            print(f"Trained model saved as {args.savemodel}")

    elif args.action == "eval":
        # Evaluate model accuracy
        if args.akida:
            model = convert(model)
            prediction_labels = model.predict(test_data.astype(np.uint8))
            accuracy = (np.sum(prediction_labels == np.argmax(test_labels, 1)) /
                        prediction_labels.shape[0])
            print(f'Akida accuracy: {accuracy}')
        else:
            evaluate_model(model,
                           test_data,
                           y=test_labels,
                           batch_size=args.batch_size)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Custom constraint for BatchNormalization layers and a method helper to apply the
constraint.
"""

import warnings

from keras.models import clone_model
from keras.utils.generic_utils import custom_object_scope
from keras.layers import BatchNormalization

from cnn2snn.min_value_constraint import MinValueConstraint


def add_gamma_constraint(model):
    """ Method helper to add a MinValueConstraint to an existing model so that
    gamma values of its BatchNormalization layers are above a defined minimum.

    This is typically used to help having a model that will be Akida compatible
    after conversion. In some cases, the mapping on hardware will fail because
    of huge values for `threshold` or `act_step` with a message indicating that
    a value cannot fit in a 20 bit signed or unsigned integer.
    In such a case, this helper can be called to apply a constraint that can fix
    the issue.

    Note that in order for the constraint to be applied to the actual weights,
    some training must be done: for an already trained model, it can be on a few
    batches, one epoch or more depending on the impact the constraint has on
    accuracy. This helper can also be called to a new model that has not been
    trained yet.

    Args:
        model (keras.Model): the model for which gamma constraints will be
            added.

    Returns:
        keras.Model: the same model with BatchNormalisation layers updated.
    """

    def apply_gamma_constraint(layer):
        constraint = MinValueConstraint()
        if isinstance(layer, BatchNormalization):
            if layer.gamma_constraint is not None:
                warnings.warn(
                    f"Layer {layer.name} already has a gamma_constraint set "
                    f"to {layer.gamma_constraint}, it will be overwritten.")
            bn = BatchNormalization.from_config(layer.get_config())
            bn.gamma_constraint = constraint
            return bn
        return layer.__class__.from_config(layer.get_config())

    with custom_object_scope({'MinValueConstraint': MinValueConstraint}):
        updated_model = clone_model(model,
                                    clone_function=apply_gamma_constraint)
    updated_model.set_weights(model.get_weights())

    return updated_model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Training script for DVS models.
"""

import os
import argparse
import numpy as np

from keras.utils.data_utils import get_file
from keras.utils.np_utils import to_categorical

from cnn2snn import load_quantized_model, convert

from ..cyclic_lr import CyclicLR
from ..training import get_training_parser, compile_model, evaluate_model

from .dvs_generator import DVS_DataGenerator


def get_data(dataset):
    """ Loads data.

    Dataset parameters for DVSDataGenerator are set from DVS pickles.

    Args:
        dataset (str): name of the dataset to load, either 'dvs_gesture' or
            'samsung_handy'.

    Returns:
        tuple: train data, test data, test labels, classes selected for training
            and number of times that every sample have to be repeated for
            training.
    """
    # Load pre-processed dataset
    train_file = get_file(fname=dataset + '_preprocessed_train.npy',
                          origin=os.path.join(
                              'http://data.brainchip.com/dataset-mirror',
                              dataset, dataset + '_preprocessed_train.npy'),
                          cache_subdir=os.path.join('datasets', dataset))

    test_file = get_file(fname=dataset + '_preprocessed_test.npy',
                         origin=os.path.join(
                             'http://data.brainchip.com/dataset-mirror',
                             dataset, dataset + '_preprocessed_test.npy'),
                         cache_subdir=os.path.join('datasets', dataset))

    train_set = list(np.load(train_file, allow_pickle=True))
    test_set = list(np.load(test_file, allow_pickle=True))

    # prepare data with DVS_DataGenerator
    if dataset == 'dvs_gesture':
        events_inversion = False
        allow_duplicate_events = False
        packets_fixed_by = 'dur'
        packets_per_sample = 3
        # packet_length is in milliseconds
        packet_length = 60
        downscale = 2
        camera_dims = (128, 128, 2)
        # class_list is used to define class that have to be learnt
        class_list = [0, 1, 2, 3, 4, 5, 6, 7, 8]
        # sample_repeats define number of time that sample should be repeated
        sample_repeats = 5
    elif dataset == 'samsung_handy':
        events_inversion = False
        allow_duplicate_events = False
        packets_fixed_by = 'dur'
        packets_per_sample = 1
        packet_length = 20
        downscale = 4
        camera_dims = (480, 640, 2)
        class_list = [0, 1, 2, 3, 4, 5, 6, 7, 8]
        sample_repeats = 30
    else:
        print('ERROR: Dataset "{}" not recognised'.format(dataset))

    # Generate DVS data from preprocessed NPY files with params defined above
    train_data = DVS_DataGenerator(
        inversion=events_inversion,
        dataset=train_set,
        subpacket_length=packet_length,
        subpackets_per_packet=packets_per_sample,
        camera_dims=camera_dims,
        packets_fixed_by=packets_fixed_by,
        include_polarity=True,
        allow_duplicate_events=allow_duplicate_events,
        downscale=downscale)

    test_data = DVS_DataGenerator(inversion=events_inversion,
                                  dataset=test_set,
                                  subpacket_length=packet_length,
                                  subpackets_per_packet=packets_per_sample,
                                  camera_dims=camera_dims,
                                  packets_fixed_by=packets_fixed_by,
                                  include_polarity=True,
                                  allow_duplicate_events=allow_duplicate_events,
                                  downscale=downscale)

    x_test_ev, y_test = test_data.get_random_samples(
        class_list=class_list, samples_per_trial=sample_repeats)
    x_test = test_data.samples_evs2images(x_test_ev)
    y_test = to_categorical(y_test)

    return train_data, x_test, y_test, class_list, sample_repeats


def train_model(model, train_set, test_set, class_list, sample_repeats,
                batch_size, max_learning_rate, epochs):
    """ Trains the model.

    Args:
        model (keras.Model): the model to train
        train_set (DVS_DataGenerator): train data generator
        test_set (tuple): test data and test labels as np.arrays
        class_list (list): classes selected for training
        sample_repeats (int): number of times that every sample have to be
         repeated for training
        batch_size (int): the batch size
        max_learning_rate (float): learning rate maximum value
        epochs (int): the number of epochs
    """
    min_lr = max_learning_rate * 0.01

    callbacks = []
    n_iterations = np.round(
        len(train_set.dataset) * sample_repeats / batch_size)

    lr_scheduler = CyclicLR(base_lr=min_lr,
                            max_lr=max_learning_rate,
                            step_size=4 * n_iterations,
                            mode='triangular2')
    callbacks.append(lr_scheduler)

    for j in range(epochs):
        print('Epoch: ' + str(j))
        x_train_ev, y_train = train_set.get_random_samples(
            class_list=class_list, samples_per_trial=sample_repeats)
        y_train = to_categorical(y_train)

        augmentation_repeats = 4
        for _ in range(augmentation_repeats):
            x_train_ev_aug = train_set.augment_data(x_train_ev,
                                                    polarity_flip=True,
                                                    rotation_range=30,
                                                    width_shift_range=1.5,
                                                    height_shift_range=1.5)

            x_train = train_set.samples_evs2images(x_train_ev_aug)
            model.fit(x_train,
                      y_train,
                      epochs=1,
                      batch_size=batch_size,
                      callbacks=callbacks,
                      validation_data=test_set)

        del x_train_ev, x_train, y_train


def main():
    """ Entry point for script and CLI usage.
    """
    global_parser = argparse.ArgumentParser(add_help=False)
    global_parser.add_argument("-d",
                               "--dataset",
                               type=str,
                               default='dvs_gesture',
                               choices=['dvs_gesture', 'samsung_handy'],
                               help="Dataset name (defaut=dvs_gesture)")

    parsers = get_training_parser(batch_size=32,
                                  tune=True,
                                  global_batch_size=False,
                                  global_parser=global_parser)

    tune_parser = parsers[2]
    tune_parser.add_argument("-ft",
                             "--fine_tune",
                             type=bool,
                             default=False,
                             help="Fine tune the model (lower learning rate)")

    args = parsers[0].parse_args()

    # Load the source model
    model = load_quantized_model(args.model)

    # Compile model
    learning_rate = 0.001
    if args.action == 'tune':
        learning_rate = 0.00001 if args.fine_tune else 0.0001

    compile_model(model, learning_rate=learning_rate)

    # Load data
    train_data, x_test, y_test, class_list, sample_repeats = get_data(
        args.dataset)

    # Train model
    if args.action in ["train", "tune"]:
        train_model(model=model,
                    train_set=train_data,
                    test_set=(x_test, y_test),
                    class_list=class_list,
                    sample_repeats=sample_repeats,
                    batch_size=args.batch_size,
                    max_learning_rate=learning_rate,
                    epochs=args.epochs)

        # Save model in Keras format (h5)
        if args.savemodel:
            model.save(args.savemodel, include_optimizer=False)
            print(f"Trained model saved as {args.savemodel}")

    elif args.action == "eval":
        # Evaluate model accuracy
        if args.akida:
            model_ak = convert(model, input_is_image=False)
            preds = model_ak.predict(np.swapaxes(x_test, 1, 2).astype('uint8'))
            accuracy = (np.argmax(y_test, 1) == preds).mean()
            print(f"Akida accuracy: {accuracy}")
        else:
            evaluate_model(model, x_test, y=y_test)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Training script for MNIST models.
"""

from keras.datasets import mnist
from keras.optimizer_v2.adam import Adam
from keras.metrics import SparseCategoricalAccuracy
from keras.losses import SparseCategoricalCrossentropy

from cnn2snn import load_quantized_model, convert

import numpy as np

from ..distiller import Distiller, KLDistillationLoss
from ..training import get_training_parser, compile_model, evaluate_model


def get_data():
    """ Loads MNIST data.

    Returns:
        tuple: train data, train labels, test data and test labels
    """
    #The data, shuffled and split between train and test sets
    (x_train, y_train), (x_test, y_test) = mnist.load_data()

    # Normalize data
    x_train = x_train.astype("float32")
    x_train = np.reshape(x_train, (-1, 28, 28, 1))

    x_test = x_test.astype("float32")
    x_test = np.reshape(x_test, (-1, 28, 28, 1))

    return x_train, y_train, x_test, y_test


def train_model(model, teacher_model, x_train, y_train, batch_size, epochs):
    """ Trains the model.

    Args:
        model (keras.Model): the model to train
        teacher_model (keras.Model): the teacher model used for the Knowledge
            distillation training
        x_train (numpy.ndarray): train data
        y_train (numpy.ndarray): train labels
        batch_size (int): the batch size
        epochs (int): the number of epochs
    """
    # Create a distiller
    distiller = Distiller(student=model, teacher=teacher_model)

    # Compile the distiller
    distiller.compile(
        optimizer=Adam(learning_rate=5e-5),
        metrics=[SparseCategoricalAccuracy()],
        student_loss_fn=SparseCategoricalCrossentropy(from_logits=True),
        distillation_loss_fn=KLDistillationLoss(temperature=10),
        alpha=0.1)

    # Train the model
    history = distiller.fit(x_train, y_train, batch_size, epochs)
    print(history.history)


def train_model_classical(model, x_train, y_train, batch_size, epochs):
    """ Trains the model without the distillation.

    Args:
        model (keras.Model): the model to train
        x_train (numpy.ndarray): train data
        y_train (numpy.ndarray): train labels
        batch_size (int): the batch size
        epochs (int): the number of epochs
    """
    history = model.fit(x_train, y_train, batch_size, epochs)
    print(history.history)


def main():
    """ Entry point for script and CLI usage.
    """
    parsers = get_training_parser(batch_size=32, global_batch_size=False)

    train_parser = parsers[1]
    train_parser.add_argument("-t",
                              "--teacher",
                              type=str,
                              default=None,
                              help="Teacher model to use for the training")

    args = parsers[0].parse_args()

    # Load the source model
    model = load_quantized_model(args.model)

    # Compile model
    compile_model(model,
                  learning_rate=5e-5,
                  loss=SparseCategoricalCrossentropy(from_logits=True))

    # Load data
    x_train, y_train, x_test, y_test = get_data()

    # Train model
    if args.action == "train":
        if args.teacher is not None:
            # Load the teacher model
            teacher_model = load_quantized_model(args.teacher)

            train_model(model, teacher_model, x_train, y_train, args.batch_size,
                        args.epochs)
        else:
            train_model_classical(model, x_train, y_train, args.batch_size,
                                  args.epochs)

        # Save Model in Keras format (h5)
        if args.savemodel:
            model.save(args.savemodel, include_optimizer=False)
            print(f"Trained model saved as {args.savemodel}")

    elif args.action == "eval":
        # Evaluate model accuracy
        if args.akida:
            model_ak = convert(model)
            preds = model_ak.predict(x_test.astype('uint8'))
            accuracy = (y_test == preds).mean()
            print(f"Accuracy : {accuracy}")
        else:
            evaluate_model(model, x_test, y=y_test)


if __name__ == "__main__":
    main()
